# future
from __future__ import annotations

# stdlib
from collections import defaultdict
from enum import Enum
from typing import Any
from typing import List
from typing import Optional
from typing import Set

# third party
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from .document_store import BaseStash
from .document_store import QueryKey
from .document_store import QueryKeys
from .document_store import StorePartition
from .response import SyftSuccess
from .serializable import serializable
from .syft_object import SyftObject


@serializable(recursive_serde=True)
class UniqueKeyCheck(Enum):
    EMPTY = 0
    MATCHES = 1
    ERROR = 2


class KeyValueBackingStore:
    def __setitem__(self, key: Any, value: Any) -> None:
        raise NotImplementedError

    def __getitem__(self, key: Any) -> Self:
        raise NotImplementedError

    def __repr__(self) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __delitem__(self, key: str):
        raise NotImplementedError

    def clear(self) -> Self:
        raise NotImplementedError

    def copy(self) -> Self:
        raise NotImplementedError

    def update(self, *args: Any, **kwargs: Any) -> Self:
        raise NotImplementedError

    def keys(self) -> Any:
        raise NotImplementedError

    def values(self) -> Any:
        raise NotImplementedError

    def items(self) -> Any:
        raise NotImplementedError

    def pop(self, *args: Any) -> Self:
        raise NotImplementedError

    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    def __iter__(self) -> Any:
        raise NotImplementedError


class KeyValueStorePartition(StorePartition):
    def init_store(self) -> None:
        super().init_store()
        self.data = self.store_config.backing_store(
            "data", self.settings, self.store_config
        )
        self.unique_keys = self.store_config.backing_store(
            "unique_keys", self.settings, self.store_config
        )
        self.searchable_keys = self.store_config.backing_store(
            "searchable_keys", self.settings, self.store_config
        )

        for partition_key in self.unique_cks:
            pk_key = partition_key.key
            if pk_key not in self.unique_keys:
                self.unique_keys[pk_key] = {}

        for partition_key in self.searchable_cks:
            pk_key = partition_key.key
            if pk_key not in self.searchable_keys:
                self.searchable_keys[pk_key] = defaultdict(list)

    def validate_partition_keys(
        self, store_query_key: QueryKey, unique_query_keys: QueryKeys
    ) -> UniqueKeyCheck:
        matches = []
        qks = unique_query_keys.all
        for qk in qks:
            pk_key, pk_value = qk.key, qk.value
            if pk_key not in self.unique_keys:
                raise Exception(
                    f"pk_key: {pk_key} not in unique_keys: {self.unique_keys.keys()}"
                )
            ck_col = self.unique_keys[pk_key]
            if pk_value in ck_col or ck_col.get(pk_value) == store_query_key.value:
                matches.append(pk_key)

        if len(matches) == 0:
            return UniqueKeyCheck.EMPTY
        elif len(matches) == len(qks):
            return UniqueKeyCheck.MATCHES

        return UniqueKeyCheck.ERROR

    def set_data_and_keys(
        self,
        store_query_key: QueryKey,
        unique_query_keys: QueryKeys,
        searchable_query_keys: QueryKeys,
        obj: SyftObject,
    ) -> None:
        # we should lock
        uqks = unique_query_keys.all

        for qk in uqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.unique_keys[pk_key]
            ck_col[pk_value] = store_query_key.value
            self.unique_keys[pk_key] = ck_col

        self.unique_keys[store_query_key.key][
            store_query_key.value
        ] = store_query_key.value

        sqks = searchable_query_keys.all
        for qk in sqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.searchable_keys[pk_key]
            if qk.type_list:
                # coerce the list of objects to strings for a single key
                pk_value = " ".join([str(obj) for obj in pk_value])

            ck_col[pk_value].append(store_query_key.value)
            self.searchable_keys[pk_key] = ck_col

        self.data[store_query_key.value] = obj

    def set(
        self, obj: SyftObject, ignore_duplicates: bool = False
    ) -> Result[SyftObject, str]:
        try:
            store_query_key = self.settings.store_key.with_obj(obj)
            exists = store_query_key.value in self.data
            unique_query_keys = self.settings.unique_keys.with_obj(obj)
            searchable_query_keys = self.settings.searchable_keys.with_obj(obj)
            ck_check = self.validate_partition_keys(
                store_query_key=store_query_key, unique_query_keys=unique_query_keys
            )
            if not exists and ck_check == UniqueKeyCheck.EMPTY:
                self.set_data_and_keys(
                    store_query_key=store_query_key,
                    unique_query_keys=unique_query_keys,
                    searchable_query_keys=searchable_query_keys,
                    obj=obj,
                )
            elif not ignore_duplicates:
                return Err(f"Duplication Key Error: {obj}")
        except Exception as e:
            return Err(f"Failed to write obj {obj}. {e}")
        return Ok(obj)

    def all(self) -> Result[List[BaseStash.object_type], str]:
        return Ok(list(self.data.values()))

    def __len__(self) -> Result[List[BaseStash.object_type], str]:
        return len(self.data)

    def find_index_or_search_keys(
        self, index_qks: QueryKeys, search_qks: QueryKeys
    ) -> Result[List[SyftObject], str]:
        ids: Optional[Set] = None
        errors = []
        if len(index_qks.all) > 0:
            index_results = self._get_keys_index(qks=index_qks)
            if index_results.is_ok():
                if ids is None:
                    ids = index_results.ok()
                ids = ids.intersection(index_results.ok())
            else:
                errors.append(index_results.err())

        search_results = None
        if len(search_qks.all) > 0:
            search_results = self._find_keys_search(qks=search_qks)

            if search_results.is_ok():
                if ids is None:
                    ids = search_results.ok()
                ids = ids.intersection(search_results.ok())
            else:
                errors.append(search_results.err())

        if len(errors) > 0:
            return Err(" ".join(errors))

        qks = self.store_query_keys(ids)
        return self.get_all_from_store(qks=qks)

    def remove_keys(
        self,
        unique_query_keys: QueryKeys,
        searchable_query_keys: QueryKeys,
    ) -> None:
        uqks = unique_query_keys.all
        for qk in uqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.unique_keys[pk_key]
            ck_col.pop(pk_value, None)

        sqks = searchable_query_keys.all
        for qk in sqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.searchable_keys[pk_key]
            ck_col.pop(pk_value, None)

    def update(self, qk: QueryKey, obj: SyftObject) -> Result[SyftObject, str]:
        try:
            if qk.value not in self.data:
                return Err(f"No object exists for query key: {qk}")

            _original_obj = self.data[qk.value]
            _original_unique_keys = self.settings.unique_keys.with_obj(_original_obj)
            _original_searchable_keys = self.settings.searchable_keys.with_obj(
                _original_obj
            )

            # 🟡 TODO 28: Add locking in this transaction

            # remove old keys
            self.remove_keys(
                unique_query_keys=_original_unique_keys,
                searchable_query_keys=_original_searchable_keys,
            )

            # update the object with new data
            for key, value in obj.to_dict(exclude_none=True).items():
                setattr(_original_obj, key, value)

            # update data and keys
            self.set_data_and_keys(
                store_query_key=qk,
                unique_query_keys=self.settings.unique_keys.with_obj(_original_obj),
                searchable_query_keys=self.settings.searchable_keys.with_obj(
                    _original_obj
                ),
                obj=_original_obj,
            )

            return Ok(_original_obj)
        except Exception as e:
            return Err(f"Failed to update obj {obj} with error: {e}")

    def get_all_from_store(self, qks: QueryKeys) -> Result[List[SyftObject], str]:
        matches = []
        for qk in qks.all:
            if qk.value in self.data:
                matches.append(self.data[qk.value])
        return Ok(matches)

    def _delete_unique_keys_for(self, obj: SyftObject) -> Result[SyftSuccess, str]:
        for _unique_ck in self.unique_cks:
            qk = _unique_ck.with_obj(obj)
            self.unique_keys[qk.key].pop(qk.value, None)
        return Ok(SyftSuccess(message="Deleted"))

    def _delete_search_keys_for(self, obj: SyftObject) -> Result[SyftSuccess, str]:
        for _search_ck in self.searchable_cks:
            qk = _search_ck.with_obj(obj)
            self.searchable_keys[qk.key].pop(qk.value, None)
        return Ok(SyftSuccess(message="Deleted"))

    def _get_keys_index(self, qks: QueryKeys) -> Result[Set[QueryKey], str]:
        try:
            # match AND
            subsets = []
            for qk in qks.all:
                subset = {}
                pk_key, pk_value = qk.key, qk.value
                if pk_key not in self.unique_keys:
                    return Err(f"Failed to query index with {qk}")
                ck_col = self.unique_keys[pk_key]
                if pk_value not in ck_col.keys():
                    # must be at least one in all query keys
                    continue
                store_value = ck_col[pk_value]
                subsets.append({store_value})

            if len(subsets) == 0:
                return Ok(set())
            # AND
            subset = subsets.pop()
            for s in subsets:
                subset = subset.intersection(s)

            return Ok(subset)
        except Exception as e:
            return Err(f"Failed to query with {qks}. {e}")

    def _find_keys_search(self, qks: QueryKeys) -> Result[Set[QueryKey], str]:
        try:
            # match AND
            subsets = []
            for qk in qks.all:
                subset = {}
                pk_key, pk_value = qk.key, qk.value
                if pk_key not in self.searchable_keys:
                    return Err(f"Failed to search with {qk}")
                ck_col = self.searchable_keys[pk_key]
                if qk.type_list:
                    # 🟡 TODO: change this hacky way to do on to many relationships
                    # this is when you search a QueryKey which is a list of items
                    # at the moment its mostly just a List[UID]
                    # match OR against all keys for this col
                    # the values of the list will be turned into strings in a single key
                    matches = set()
                    for item in pk_value:
                        for col_key in ck_col.keys():
                            if str(item) in col_key:
                                store_values = ck_col[col_key]
                                for value in store_values:
                                    matches.add(value)
                    if len(matches):
                        subsets.append(matches)
                else:
                    # this is the normal path
                    if pk_value not in ck_col.keys():
                        # must be at least one in all query keys
                        continue
                    store_values = ck_col[pk_value]
                    subsets.append(set(store_values))

            if len(subsets) == 0:
                return Ok(set())
            # AND
            subset = subsets.pop()
            for s in subsets:
                subset = subset.intersection(s)
            return Ok(subset)
        except Exception as e:
            return Err(f"Failed to query with {qks}. {e}")

    def create(self, obj: SyftObject) -> Result[SyftObject, str]:
        pass

    def delete(self, qk: QueryKey) -> Result[SyftSuccess, Err]:
        try:
            _obj = self.data.pop(qk.value)
            self._delete_unique_keys_for(_obj)
            self._delete_search_keys_for(_obj)
            return Ok(SyftSuccess(message="Deleted"))
        except Exception as e:
            return Err(f"Failed to delete with query key {qk} with error: {e}")



# stdlib
from typing import Any
from typing import ClassVar
from typing import Type

# third party
import numpy as np

# relative
from .action_object import ActionObject
from .action_types import action_types
from .serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1

# @serializable(recursive_serde=True)
# class NumpyArrayObjectPointer(ActionObjectPointer):
#     _inflix_operations = ["__add__", "__sub__", "__eq__", "__mul__"]
#     __canonical_name__ = "NumpyArrayObjectPointer"
#     __version__ = SYFT_OBJECT_VERSION_1

#     # 🟡 TODO 17: add state / allowlist inheritance to SyftObject and ignore methods by default
#     __attr_state__ = [
#         "id",
#         "node_uid",
#         "parent_id",
#     ]

#     def get_from(self, domain_client) -> Any:
#         return domain_client.api.services.action.get(self.id).syft_action_data


class NumpyArrayObjectPointer:
    pass


def numpy_like_eq(left: Any, right: Any) -> bool:
    result = left == right
    if isinstance(result, bool):
        return result

    if hasattr(result, "all"):
        return (result).all()
    return bool(result)


# 🔵 TODO 7: Map TPActionObjects and their 3rd Party types like numpy type to these
# classes for bi-directional lookup.
@serializable(recursive_serde=True)
class NumpyArrayObject(ActionObject, np.lib.mixins.NDArrayOperatorsMixin):
    __canonical_name__ = "NumpyArrayObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: ClassVar[Type[Any]] = np.ndarray
    syft_pointer_type = NumpyArrayObjectPointer
    syft_passthrough_attrs = []
    syft_dont_wrap_attrs = ["dtype"]

    # def __eq__(self, other: Any) -> bool:
    #     # 🟡 TODO 8: move __eq__ to a Data / Serdeable type interface on ActionObject
    #     if isinstance(other, NumpyArrayObject):
    #         return (
    #             numpy_like_eq(self.syft_action_data, other.syft_action_data)
    #             and self.syft_pointer_type == other.syft_pointer_type
    #         )
    #     return self == other

    def __bool__(self) -> bool:
        return bool(self.all())

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        inputs = tuple(
            np.array(x.syft_action_data, dtype=x.dtype)
            if isinstance(x, NumpyArrayObject)
            else x
            for x in inputs
        )

        result = getattr(ufunc, method)(*inputs, **kwargs)
        if type(result) is tuple:
            return tuple(
                NumpyArrayObject(syft_action_data=x, dtype=x.dtype, shape=x.shape)
                for x in result
            )
        else:
            return NumpyArrayObject(
                syft_action_data=result, dtype=result.dtype, shape=result.shape
            )


@serializable(recursive_serde=True)
class NumpyScalarObject(ActionObject, np.lib.mixins.NDArrayOperatorsMixin):
    __canonical_name__ = "NumpyScalarObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type = np.number
    syft_passthrough_attrs = []
    syft_dont_wrap_attrs = ["dtype"]

    def __float__(self) -> float:
        return float(self.syft_action_data)


@serializable(recursive_serde=True)
class NumpyBoolObject(ActionObject, np.lib.mixins.NDArrayOperatorsMixin):
    __canonical_name__ = "NumpyBoolObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type = np.bool_
    syft_passthrough_attrs = []
    syft_dont_wrap_attrs = ["dtype"]


np_array = np.array([1, 2, 3])
action_types[type(np_array)] = NumpyArrayObject


SUPPORTED_BOOL_TYPES = [np.bool_]

for scalar_type in SUPPORTED_BOOL_TYPES:
    action_types[scalar_type] = NumpyBoolObject


SUPPORTED_INT_TYPES = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

SUPPORTED_FLOAT_TYPES = [
    np.float16,
    np.float32,
    np.float64,
]

for scalar_type in SUPPORTED_INT_TYPES + SUPPORTED_FLOAT_TYPES:
    action_types[scalar_type] = NumpyScalarObject



# stdlib
from typing import List
from typing import Union

# relative
from ....telemetry import instrument
from .context import AuthedServiceContext
from .dataset import Asset
from .dataset import CreateDataset
from .dataset import Dataset
from .dataset_stash import DatasetStash
from .document_store import DocumentStore
from .response import SyftError
from .response import SyftSuccess
from .serializable import serializable
from .service import AbstractService
from .service import SERVICE_TO_TYPES
from .service import TYPE_TO_SERVICE
from .service import service_method
from .uid import UID


@instrument
@serializable(recursive_serde=True)
class DatasetService(AbstractService):
    store: DocumentStore
    stash: DatasetStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = DatasetStash(store=store)

    @service_method(path="dataset.add", name="add")
    def add(
        self, context: AuthedServiceContext, dataset: CreateDataset
    ) -> Union[SyftSuccess, SyftError]:
        """Add a Dataset"""
        result = self.stash.set(dataset.to(Dataset, context=context))
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Dataset Added")

    @service_method(path="dataset.get_all", name="get_all")
    def get_all(self, context: AuthedServiceContext) -> Union[List[Dataset], SyftError]:
        """Get a Dataset"""
        result = self.stash.get_all()
        if result.is_ok():
            datasets = result.ok()
            results = []
            for dataset in datasets:
                dataset.node_uid = context.node.id
                results.append(dataset)
            return results
        return SyftError(message=result.err())

    @service_method(path="dataset.search", name="search")
    def search(
        self, context: AuthedServiceContext, name: str
    ) -> Union[List[Dataset], SyftError]:
        """Search a Dataset by name"""
        results = self.get_all(context)

        return (
            results
            if isinstance(results, SyftError)
            else [dataset for dataset in results if name in dataset.name]
        )

    @service_method(path="dataset.get_by_id", name="get_by_id")
    def get_by_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        """Get a Dataset"""
        result = self.stash.get_by_uid(uid=uid)
        if result.is_ok():
            dataset = result.ok()
            dataset.node_uid = context.node.id
            return dataset
        return SyftError(message=result.err())

    @service_method(path="dataset.get_by_action_id", name="get_by_action_id")
    def get_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Dataset], SyftError]:
        """Get Datasets by an Action ID"""
        result = self.stash.search_action_ids(uid=uid)
        if result.is_ok():
            datasets = result.ok()
            for dataset in datasets:
                dataset.node_uid = context.node.id
            return datasets
        return SyftError(message=result.err())

    @service_method(
        path="dataset.get_assets_by_action_id", name="get_assets_by_action_id"
    )
    def get_assets_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Asset], SyftError]:
        """Get Assets by an Action ID"""
        datasets = self.get_by_action_id(context=context, uid=uid)
        assets = []
        if isinstance(datasets, list):
            for dataset in datasets:
                for asset in dataset.asset_list:
                    if asset.action_id == uid:
                        assets.append(asset)
            return assets
        elif isinstance(datasets, SyftError):
            return datasets
        return []


TYPE_TO_SERVICE[Dataset] = DatasetService
SERVICE_TO_TYPES[DatasetService].update({Dataset})



# stdlib
from typing import List
from typing import Union

# relative
from ....telemetry import instrument
from .context import AuthedServiceContext
from .document_store import DocumentStore
from .message_stash import MessageStash
from .messages import CreateMessage
from .messages import LinkedObject
from .messages import Message
from .messages import MessageStatus
from .response import SyftError
from .response import SyftSuccess
from .serializable import serializable
from .service import AbstractService
from .service import SERVICE_TO_TYPES
from .service import TYPE_TO_SERVICE
from .service import service_method
from .uid import UID


@instrument
@serializable(recursive_serde=True)
class MessageService(AbstractService):
    store: DocumentStore
    stash: MessageStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = MessageStash(store=store)

    @service_method(path="messages.send", name="send")
    def send(
        self, context: AuthedServiceContext, message: CreateMessage
    ) -> Union[Message, SyftError]:
        """Send a new message"""

        new_message = message.to(Message, context=context)
        result = self.stash.set(new_message)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="messages.get_all", name="get_all")
    def get_all(self, context: AuthedServiceContext) -> Union[List[Message], SyftError]:
        result = self.stash.get_all_inbox_for_verify_key(verify_key=context.credentials)
        if result.err():
            return SyftError(message=str(result.err()))
        messages = result.ok()
        return messages

    @service_method(path="messages.get_all_sent", name="outbox")
    def get_all_sent(
        self, context: AuthedServiceContext
    ) -> Union[List[Message], SyftError]:
        result = self.stash.get_all_sent_for_verify_key(verify_key=context.credentials)
        if result.err():
            return SyftError(message=str(result.err()))
        messages = result.ok()
        print(messages)
        return messages

    @service_method(
        path="messages.get_all_for_status",
        name="get_all_for_status",
    )
    def get_all_for_status(
        self,
        context: AuthedServiceContext,
        status: MessageStatus,
    ) -> Union[List[Message], SyftError]:
        result = self.stash.get_all_by_verify_key_for_status(
            verify_key=context.credentials, status=status
        )
        if result.err():
            return SyftError(message=str(result.err()))
        messages = result.ok()
        return messages

    @service_method(path="messages.mark_as_delivered", name="mark_as_delivered")
    def mark_as_delivered(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[Message, SyftError]:
        result = self.stash.update_message_status(
            uid=uid, status=MessageStatus.DELIVERED
        )
        if result.is_err():
            return SyftError(message=str(result.err()))

        return result.ok()

    @service_method(path="messages.resolve_object", name="resolve_object")
    def resolve_object(
        self, context: AuthedServiceContext, linked_obj: LinkedObject
    ) -> Union[Message, SyftError]:
        service = context.node.get_service(linked_obj.service_type)
        result = service.resolve_link(context=context, linked_obj=linked_obj)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="messages.clear", name="clear")
    def clear(self, context: AuthedServiceContext) -> Union[SyftError, SyftSuccess]:
        result = self.stash.delete_all_for_verify_key(verify_key=context.credentials)
        if result.is_ok():
            return SyftSuccess(message="All messages cleared !!")
        return SyftError(message=str(result.err()))


TYPE_TO_SERVICE[Message] = MessageService
SERVICE_TO_TYPES[MessageService].update({Message})



# future
from __future__ import annotations

# stdlib
import inspect
import types
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import KeysView
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union

# third party
import pydantic
from typing_extensions import Self

# relative
from .action_data_empty import ActionDataEmpty
from .action_types import action_type_for_type
from .action_types import action_types
from .client import SyftClient
from .response import SyftException
from .serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftBaseObject
from .syft_object import SyftObject
from .uid import LineageID
from .uid import UID


@serializable(recursive_serde=True)
class Action(SyftObject):
    __canonical_name__ = "Action"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__: List[str] = []

    path: str
    op: str
    remote_self: Optional[LineageID]
    args: List[LineageID]
    kwargs: Dict[str, LineageID]
    result_id: Optional[LineageID]

    @pydantic.validator("id", pre=True, always=True)
    def make_id(cls, v: Optional[UID]) -> UID:
        return v if isinstance(v, UID) else UID()

    @pydantic.validator("result_id", pre=True, always=True)
    def make_result_id(cls, v: Optional[Union[UID, LineageID]]) -> UID:
        return v if isinstance(v, LineageID) else LineageID(v)

    @property
    def full_path(self) -> str:
        return f"{self.path}.{self.op}"

    @property
    def syft_history_hash(self) -> int:
        hashes = 0
        hashes += hash(self.remote_self.syft_history_hash)
        # 🔵 TODO: resolve this
        # if the object is ActionDataEmpty then the type might not be equal to the
        # real thing. This is the same issue with determining the result type from
        # a pointer operation in the past, so we should think about what we want here
        # hashes += hash(self.path)
        hashes += hash(self.op)
        for arg in self.args:
            hashes += hash(arg.syft_history_hash)
        for k, arg in self.kwargs:
            hashes += hash(k)
            hashes += hash(arg.syft_history_hash)
        return hashes


def __make_infix_op__(op: str) -> Callable:
    def infix_op(_self, other: Any) -> Self:
        if not isinstance(other, ActionObjectPointer):
            other = other.to_pointer(_self.node_uid)

        action = _self.syft_make_method_action(op=op, args=[other])
        action_result = _self.syft_execute_action(action, sync=True)
        return action_result

    infix_op.__name__ = op
    return infix_op


class ActionObjectPointer:
    pass


# class ActionObjectPointer(SyftObject):
#     __canonical_name__ = "ActionObjectPointer"
#     __version__ = SYFT_OBJECT_VERSION_1

#     __attr_state__ = ["id", "node_uid", "parent_id"]

#     _inflix_operations: List = []

#     node_uid: Optional[UID]
#     parent_id: Optional[UID]

#     def __new__(cls, *args, **kwargs):
#         for op in cls._inflix_operations:
#             new_op = __make_infix_op__(op)
#             setattr(ActionObjectPointer, op, new_op)
#         return super(ActionObjectPointer, cls).__new__(cls)

HOOK_ALWAYS = "ALWAYS"

passthrough_attrs = [
    "__dict__",  # python
    "__class__",  # python
    "__repr_name__",  # python
    "__annotations__",  # python
    "_init_private_attributes",  # pydantic
    "__private_attributes__",  # pydantic
    "__config__",  # pydantic
    "__fields__",  # pydantic
    "__fields_set__",  # pydantic
    "__repr_str__",  # pydantic
    "__repr_args__",  # pydantic
    "__post_init__",  # syft
    "id",  # syft
    "to_mongo",  # syft 🟡 TODO 23: Add composeable / inheritable object passthrough attrs
    "__attr_searchable__",  # syft
    "__canonical_name__",  # syft
    "__version__",  # syft
    "__args__",  # pydantic
    "to_pointer",  # syft
    "to",  # syft
    "send",  # syft
    "_copy_and_set_values",  # pydantic
    "get_from",  # syft
]
dont_wrap_output_attrs = [
    "__repr__",
    "__str__",
    "_repr_html_",
    "_repr_markdown_",
    "_repr_latex_",
    "__array_struct__",
    "__array_prepare__",
    "__array_wrap__",
    "__bool__",
    "__len__",
]
dont_make_side_effects = [
    "_repr_html_",
    "_repr_markdown_",
    "_repr_latex_",
    "__repr__",
    "__getitem__",
    "__setitem__",
    "__len__",
    "shape",
]
action_data_empty_must_run = [
    "__repr__",
]
show_print = False


def debug_original_func(name, func):
    print(f"{name} func is:")
    print("inspect.isdatadescriptor", inspect.isdatadescriptor(func))
    print("inspect.isgetsetdescriptor", inspect.isgetsetdescriptor(func))
    print("inspect.isfunction", inspect.isfunction(func))
    print("inspect.isbuiltin", inspect.isbuiltin(func))
    print("inspect.ismethod", inspect.ismethod(func))
    print("inspect.ismethoddescriptor", inspect.ismethoddescriptor(func))


def make_action_side_effect(context: PreHookContext, *args: Any, **kwargs: Any) -> Any:
    try:
        action = context.obj.syft_make_method_action(
            op=context.op_name, args=args, kwargs=kwargs
        )
        context.action = action
    except Exception as e:
        print(
            "Exception detected in make_action_side_effect", e
        )  # TODO: Put this Exception back
        # pass
    return context, args, kwargs


def send_action_side_effect(context: PreHookContext, *args: Any, **kwargs: Any) -> Any:
    try:
        if context.op_name not in dont_make_side_effects and hasattr(
            context.obj, "syft_node_uid"
        ):
            if getattr(context.obj, "syft_node_uid", None):
                if context.action is not None:
                    action = context.obj.syft_make_method_action(
                        op=context.op_name, args=args, kwargs=kwargs
                    )
                    context.action = action

                action_result = context.obj.syft_execute_action(action, sync=True)
                if not isinstance(action_result, ActionObject):
                    print("Got back unexpected response", action_result)
                else:
                    context.node_uid = action_result.syft_node_uid
                    context.result_id = action.result_id
                    print("IGNORING: got action result", action_result)
            else:
                # 🟡 TODO
                pass
                # print(
                #     "Can't Send Action without a target node. Use .point_to(node_uid: UID)"
                # )
    except Exception as e:
        print(
            "Exception in send_action_side_effect", e
        )  # TODO: Put this Exception back
        # pass
    return context, args, kwargs


def propagate_node_uid(context: PreHookContext, op: str, result: Any) -> Any:
    try:
        if context.op_name not in dont_make_side_effects and hasattr(
            context.obj, "syft_node_uid"
        ):
            syft_node_uid = getattr(context.obj, "syft_node_uid", None)
            if syft_node_uid:
                if not hasattr(result, "syft_node_uid"):
                    # print("result doesnt have a syft_node_uid attr")
                    pass
                if op not in context.obj._syft_dont_wrap_attrs():
                    if hasattr(result, "syft_node_uid"):
                        setattr(result, "syft_node_uid", syft_node_uid)
                else:
                    print("dont propogate node_uid because output isnt wrapped")
            else:
                # 🟡 TODO
                # print("Can't proagate node_uid because parent doesnt have one")
                pass
    except Exception as e:
        print("Exception in propagate_node_uid", e)
    return result


def is_action_data_empty(obj: Any) -> bool:
    if hasattr(obj, "syft_action_data"):
        obj = obj.syft_action_data
    if isinstance(obj, ActionDataEmpty):
        return True
    return False


def has_action_data_empty(args: Any, kwargs: Any) -> bool:
    for a in args:
        if is_action_data_empty(a):
            return True

    for _, a in kwargs.items():
        if is_action_data_empty(a):
            return True
    return False


def debox_args_and_kwargs(args: Any, kwargs: Any) -> Tuple[Any, Any]:
    filtered_args = []
    filtered_kwargs = {}
    for a in args:
        value = a
        if hasattr(value, "syft_action_data"):
            value = value.syft_action_data
        filtered_args.append(value)

    for k, a in kwargs.items():
        value = a
        if hasattr(value, "syft_action_data"):
            value = value.syft_action_data
        filtered_kwargs[k] = a

    return tuple(filtered_args), filtered_kwargs


class PreHookContext(SyftBaseObject):
    obj: Any
    op_name: str
    node_uid: Optional[UID]
    result_id: Optional[Union[UID, LineageID]]
    action: Optional[Action]


class ActionObject(SyftObject):
    __attr_searchable__: List[str] = []
    __canonical_name__ = "ActionObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_action_data: Optional[Any] = None
    syft_pointer_type: ClassVar[Type[ActionObjectPointer]]

    # Help with calculating history hash for code verification
    syft_parent_hashes: Optional[Union[int, List[int]]]
    syft_parent_op: Optional[str]
    syft_parent_args: Optional[Any]
    syft_parent_kwargs: Optional[Any]
    syft_history_hash: Optional[int]
    syft_internal_type: ClassVar[Type[Any]]

    @property
    def syft_lineage_id(self) -> LineageID:
        return LineageID(self.id, self.syft_history_hash)

    @pydantic.validator("id", pre=True, always=True)
    def make_id(cls, v: Optional[UID]) -> UID:
        return v if isinstance(v, UID) else UID()

    @pydantic.validator("syft_action_data", pre=True, always=True)
    def check_action_data(
        cls, v: ActionObject.syft_pointer_type
    ) -> ActionObject.syft_pointer_type:
        if cls == AnyActionObject or isinstance(
            v, (cls.syft_internal_type, ActionDataEmpty)
        ):
            return v
        raise SyftException(
            f"Must init {cls} with {cls.syft_internal_type} not {type(v)}"
        )

    def syft_point_to(self, node_uid: UID) -> None:
        self.syft_node_uid = node_uid

    def syft_get_property(self, obj, method) -> Any:
        klass_method = getattr(type(obj), method, None)
        return klass_method.__get__(obj)

    def syft_is_property(self, obj, method) -> bool:
        klass_method = getattr(type(obj), method, None)
        return isinstance(klass_method, property) or inspect.isdatadescriptor(
            klass_method
        )

    def send(self, client: SyftClient) -> Self:
        return client.api.services.action.set(self)

    def get_from(self, domain_client) -> Any:
        return domain_client.api.services.action.get(self.id).syft_action_data

    @staticmethod
    def from_obj(
        syft_action_data: Any,
        id: Optional[UID] = None,
        syft_lineage_id: Optional[LineageID] = None,
    ) -> ActionObject:
        if id and syft_lineage_id and id != syft_lineage_id.id:
            raise Exception("UID and LineageID should match")
        action_type = action_type_for_type(syft_action_data)
        if action_type is None:
            raise Exception(f"{type(syft_action_data)} not in action_types")
        action_object = action_type(syft_action_data=syft_action_data)
        if id:
            action_object.id = id
        if syft_lineage_id:
            action_object.id = syft_lineage_id.id
            action_object.syft_history_hash = syft_lineage_id.syft_history_hash
        elif id:
            action_object.syft_history_hash = hash(id)

        return action_object

    @staticmethod
    def empty(
        syft_internal_type: Optional[Any] = Any,
        id: Optional[UID] = None,
        syft_lineage_id: Optional[LineageID] = None,
    ) -> ActionObject:
        empty = ActionDataEmpty(syft_internal_type=syft_internal_type)
        action_object = ActionObject.from_obj(
            syft_action_data=empty, id=id, syft_lineage_id=syft_lineage_id
        )
        return action_object

    syft_action_data: Union[Any, Tuple[Any, Any]]
    syft_node_uid: Optional[UID]
    #  = (
    #     PrivateAttr()
    # )  # 🔵 TODO 6: Make special ActionObject attrs _syft if possible
    _syft_pre_hooks__: Dict[str, Set] = {}
    _syft_post_hooks__: Dict[str, Set] = {}

    class Config:
        arbitrary_types_allowed = True

    def __post_init__(self) -> None:
        if HOOK_ALWAYS not in self._syft_pre_hooks__:
            self._syft_pre_hooks__[HOOK_ALWAYS] = set()
        self._syft_pre_hooks__[HOOK_ALWAYS].add(make_action_side_effect)
        self._syft_pre_hooks__[HOOK_ALWAYS].add(send_action_side_effect)

        if HOOK_ALWAYS not in self._syft_post_hooks__:
            self._syft_post_hooks__[HOOK_ALWAYS] = set()
        self._syft_post_hooks__[HOOK_ALWAYS].add(propagate_node_uid)

        if isinstance(self.syft_action_data, ActionObject):
            raise Exception("Nested ActionObjects", self.syft_action_data)

        self.syft_history_hash = hash(self.id)

    def _syft_run_pre_hooks__(
        self, context, name, args, kwargs
    ) -> Tuple[PreHookContext, Tuple[Any, ...], Dict[str, Any]]:
        try:
            result_args, result_kwargs = args, kwargs
            if name in self._syft_pre_hooks__:
                for hook in self._syft_pre_hooks__[name]:
                    context, result_args, result_kwargs = hook(
                        context, *result_args, **result_kwargs
                    )

            if name not in self._syft_dont_wrap_attrs():
                if HOOK_ALWAYS in self._syft_pre_hooks__:
                    for hook in self._syft_pre_hooks__[HOOK_ALWAYS]:
                        context, result_args, result_kwargs = hook(
                            context, *result_args, **result_kwargs
                        )
        except Exception as e:
            print("Exception in pre hooks", e)
        return context, result_args, result_kwargs

    def _syft_run_post_hooks__(self, context, name, result):
        new_result = result
        if name in self._syft_post_hooks__:
            for hook in self._syft_post_hooks__[name]:
                new_result = hook(context, name, new_result)

        if name not in self._syft_dont_wrap_attrs():
            if HOOK_ALWAYS in self._syft_post_hooks__:
                for hook in self._syft_post_hooks__[HOOK_ALWAYS]:
                    new_result = hook(context, name, new_result)
        return new_result

    def _syft_output_action_object(
        self,
        result,
    ) -> Any:
        # can check types here
        if not issubclass(type(result), ActionObject):
            constructor = action_type_for_type(result)
            if not constructor:
                raise Exception(f"output: {type(result)} no in action_types")
            result = constructor(syft_action_data=result)

        return result

    def _syft_passthrough_attrs(self) -> List[str]:
        return passthrough_attrs + getattr(self, "syft_passthrough_attrs", [])

    def _syft_dont_wrap_attrs(self) -> List[str]:
        return dont_wrap_output_attrs + getattr(self, "syft_dont_wrap_attrs", [])

    def __getattribute__(self, name):
        # bypass certain attrs to prevent recursion issues
        if name.startswith("_syft") or name.startswith("syft"):
            return object.__getattribute__(self, name)

        if name in self._syft_passthrough_attrs():
            return object.__getattribute__(self, name)
        defined_on_self = name in self.__dict__ or name in self.__private_attributes__
        if show_print:
            print(">> ", name, ", defined_on_self = ", defined_on_self)

        # use the custom definied version
        context_self = self
        if not defined_on_self:
            context_self = self.syft_action_data

        # TODO: weird edge cases here for things like tuples
        if name == "__bool__" and not hasattr(self.syft_action_data, "__bool__"):
            context = PreHookContext(obj=self, op_name=name)
            context, _, _ = self._syft_run_pre_hooks__(context, name, (), {})
            # no input needs to propagate
            result = self._syft_run_post_hooks__(
                context, name, bool(self.syft_action_data)
            )
            if name not in self._syft_dont_wrap_attrs():
                result = self._syft_output_action_object(result)
                if context.action is not None:
                    result.syft_history_hash = context.action.syft_history_hash

            def __wrapper__bool__() -> bool:
                return result

            return __wrapper__bool__

        if self.syft_is_property(context_self, name):
            if show_print:
                print("Property detected: ", name)
            if self.syft_is_property(context_self, name):
                context = PreHookContext(obj=self, op_name=name)
                context, _, _ = self._syft_run_pre_hooks__(context, name, (), {})
                # no input needs to propagate
                result = self._syft_run_post_hooks__(
                    context, name, self.syft_get_property(context_self, name)
                )
                if name not in self._syft_dont_wrap_attrs():
                    result = self._syft_output_action_object(result)
                    if context.action is not None:
                        result.syft_history_hash = context.action.syft_history_hash
                return result

        # check for other types that aren't methods, functions etc
        def fake_func(*args, **kwargs) -> Any:
            return ActionDataEmpty(syft_internal_type=self.syft_internal_type)

        if (
            isinstance(self.syft_action_data, ActionDataEmpty)
            and name not in action_data_empty_must_run
        ):
            original_func = fake_func
        else:
            original_func = getattr(self.syft_action_data, name)

        if show_print:
            debug_original_func(name, original_func)
        if inspect.ismethod(original_func) or inspect.ismethoddescriptor(original_func):
            if show_print:
                print("Running method: ", name)

            def wrapper(_self, *args, **kwargs):
                context = PreHookContext(obj=self, op_name=name)
                context, pre_hook_args, pre_hook_kwargs = self._syft_run_pre_hooks__(
                    context, name, args, kwargs
                )

                if not has_action_data_empty(args=args, kwargs=kwargs):
                    original_args, original_kwargs = debox_args_and_kwargs(
                        pre_hook_args, pre_hook_kwargs
                    )

                    result = original_func(*original_args, **original_kwargs)
                else:
                    result = fake_func(*args, **kwargs)

                post_result = self._syft_run_post_hooks__(context, name, result)
                if name not in self._syft_dont_wrap_attrs():
                    post_result = self._syft_output_action_object(
                        post_result,
                    )
                    if context.action is not None:
                        post_result.syft_history_hash = context.action.syft_history_hash
                    post_result.syft_node_uid = context.node_uid
                    if context.result_id is not None:
                        post_result.id = context.result_id
                return post_result

            wrapper = types.MethodType(wrapper, type(self))
        else:
            if show_print:
                print("Running non-method: ", name)

            def wrapper(*args, **kwargs):
                context = PreHookContext(obj=self, op_name=name)
                context, pre_hook_args, pre_hook_kwargs = self._syft_run_pre_hooks__(
                    context, name, args, kwargs
                )

                if not has_action_data_empty(args=args, kwargs=kwargs):
                    original_args, original_kwargs = debox_args_and_kwargs(
                        pre_hook_args, pre_hook_kwargs
                    )

                    result = original_func(*original_args, **original_kwargs)
                else:
                    result = fake_func(*args, **kwargs)

                post_result = self._syft_run_post_hooks__(context, name, result)
                if name not in self._syft_dont_wrap_attrs():
                    post_result = self._syft_output_action_object(post_result)
                    if context.action is not None:
                        post_result.syft_history_hash = context.action.syft_history_hash
                    post_result.syft_node_uid = context.node_uid
                    if context.result_id is not None:
                        post_result.id = context.result_id
                return post_result

        try:
            wrapper.__doc__ = original_func.__doc__
            if show_print:
                print(
                    "Found original signature for ",
                    name,
                    inspect.signature(original_func),
                )
            wrapper.__ipython_inspector_signature_override__ = inspect.signature(
                original_func
            )
        except Exception:
            if show_print:
                print("name", name, "has no signature")

        return wrapper

    def syft_execute_action(
        self, action: Action, sync: bool = True
    ) -> ActionObjectPointer:
        if self.syft_node_uid is None:
            raise SyftException("Pointers can't execute without a node_uid.")
        # relative
        from .api import APIRegistry
        from .api import SyftAPICall

        api = APIRegistry.api_for(node_uid=self.syft_node_uid)

        kwargs = {"action": action}
        api_call = SyftAPICall(
            node_uid=self.syft_node_uid, path="action.execute", args=[], kwargs=kwargs
        )
        return api.make_call(api_call)

    def syft_make_action(
        self,
        path: str,
        op: str,
        remote_self: Optional[Union[UID, LineageID]] = None,
        args: Optional[List[Union[UID, ActionObjectPointer, LineageID]]] = None,
        kwargs: Optional[Dict[str, Union[UID, ActionObjectPointer, LineageID]]] = None,
    ) -> Action:
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        arg_ids = [
            LineageID(uid) if isinstance(uid, (UID, LineageID)) else uid.syft_lineage_id
            for uid in args
        ]
        kwarg_ids = {
            k: LineageID(uid)
            if isinstance(uid, (LineageID, UID))
            else uid.syft_lineage_id
            for k, uid in kwargs.items()
        }
        action = Action(
            path=path,
            op=op,
            remote_self=LineageID(remote_self),
            args=arg_ids,
            kwargs=kwarg_ids,
        )
        return action

    def syft_make_method_action(
        self,
        op: str,
        args: Optional[List[Union[UID, ActionObjectPointer]]] = None,
        kwargs: Optional[Dict[str, Union[UID, ActionObjectPointer]]] = None,
    ) -> Action:
        path = self.syft_get_path()
        return self.syft_make_action(
            path=path, op=op, remote_self=self.syft_lineage_id, args=args, kwargs=kwargs
        )

    def syft_get_path(self) -> str:
        if isinstance(self, AnyActionObject) and self.syft_internal_type:
            return f"{self.syft_internal_type.__name__}"
        return f"{type(self).__name__}"

    def syft_remote_method(
        self,
        op: str,
    ) -> Action:
        def wrapper(
            *args: Optional[List[Union[UID, ActionObjectPointer]]],
            **kwargs: Optional[Dict[str, Union[UID, ActionObjectPointer]]],
        ) -> Action:
            return self.syft_make_method_action(op=op, args=args, kwargs=kwargs)

        return wrapper

    def keys(self) -> KeysView[str]:
        return self.syft_action_data.keys()

    ###### __DUNDER_MIFFLIN__

    # if we do not implement these boiler plate __method__'s then special infix
    # operations like x + y won't trigger __getattribute__
    # unless there is a super special reason we should write no code in these functions

    def __repr__(self) -> str:
        return self.__repr__()

    def __str__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return self.__len__()

    def __getitem__(self, key: Any) -> Any:
        return self._syft_output_action_object(self.__getitem__(key))

    def __setitem__(self, key: Any, value: Any) -> None:
        return self.__setitem__(key, value)

    def __contains__(self, key: Any) -> bool:
        return self.__contains__(key)

    def __bool__(self) -> bool:
        return self.__bool__()

    def __add__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__add__(other))

    def __sub__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__sub__(other))

    def __mul__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__mul__(other))

    def __matmul__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__matmul__(other))

    def __eq__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__eq__(other))

    def __lt__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__lt__(other))

    def __gt__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__gt__(other))

    def __le__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__le__(other))

    def __ge__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__ge__(other))

    def __delattr__(self, key: Any) -> None:
        self.__delattr__(key)

    def __delitem__(self, key: Any) -> None:
        self.__delitem__(key)

    def __invert__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__invert__(other))

    def __divmod__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__divmod__(other))

    def __floordiv__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__floordiv__(other))

    def __mod__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__mod__(other))

    # r ops
    # we want the underlying implementation so we should just call into __getattribute__
    def __radd__(self, other: Any) -> Any:
        return self.__radd__(other)

    def __rsub__(self, other: Any) -> Any:
        return self.__rsub__(other)

    def __rmatmul__(self, other: Any) -> Any:
        return self.__rmatmul__(other)


@serializable(recursive_serde=True)
class AnyActionObject(ActionObject):
    __canonical_name__ = "AnyActionObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: ClassVar[Type[Any]] = Any
    syft_passthrough_attrs = []
    syft_dont_wrap_attrs = []

    def __float__(self) -> float:
        return float(self.syft_action_data)


action_types[Any] = AnyActionObject



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional
from typing import Type

# relative
from .serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject


@serializable(recursive_serde=True)
class ActionDataEmpty(SyftObject):
    __canonical_name__ = "ActionDataEmpty"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: Optional[Type] = Any



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional
from typing import Type

# relative
from .document_store import DocumentStore
from .document_store import StoreConfig
from .kv_document_store import KeyValueBackingStore
from .kv_document_store import KeyValueStorePartition
from .serializable import serializable


@serializable(recursive_serde=True)
class DictBackingStore(dict, KeyValueBackingStore):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(dict).__init__()
        self._ddtype = kwargs.get("ddtype", None)

    def __getitem__(self, key: Any) -> Any:
        try:
            value = super().__getitem__(key)
            return value
        except KeyError as e:
            if self._ddtype:
                return self._ddtype()
            raise e


@serializable(recursive_serde=True)
class DictStorePartition(KeyValueStorePartition):
    def prune(self):
        self.init_store()


# the base document store is already a dict but we can change it later
@serializable(recursive_serde=True)
class DictDocumentStore(DocumentStore):
    partition_type = DictStorePartition

    def __init__(self, store_config: Optional[DictStoreConfig] = None) -> None:
        if store_config is None:
            store_config = DictStoreConfig()
        super().__init__(store_config=store_config)

    def reset(self):
        for _, partition in self.partitions.items():
            partition.prune()


@serializable(recursive_serde=True)
class DictStoreConfig(StoreConfig):
    store_type: Type[DocumentStore] = DictDocumentStore
    backing_store: Type[KeyValueBackingStore] = DictBackingStore



# stdlib
import base64
import os
import platform
import subprocess  # nosec
import sys
import tarfile
import zipfile

# third party
import requests

# relative
from ...util import bcolors


def check_oblv_proxy_installation_status():
    try:
        result = subprocess.run(["oblv", "-V"], capture_output=True, text=True)  # nosec
        if result.stderr:
            raise subprocess.CalledProcessError(  # nosec
                returncode=result.returncode, cmd=result.args, stderr=result.stderr
            )
        result = result.stdout.strip()
        return result
    except Exception as e:
        if e.__class__ == FileNotFoundError:
            system_name = platform.system()
            result = "Oblv Proxy Not Installed. Call the method install_oblv_proxy "
            if system_name == "Windows":
                result += "to install the proxy for this session. If you already have the proxy installed,"
                +" add it to your PATH."
            elif system_name == "Linux":
                result += "to install the proxy globally. If you already have the proxy installed,"
                +" create a link to the installation as /usr/local/bin/oblv"

            print(
                bcolors.RED
                + bcolors.BOLD
                + "Exception"
                + bcolors.BLACK
                + bcolors.ENDC
                + ": "
                + result,
                file=sys.stderr,
            )
        else:
            raise Exception(e)


def install_oblv_proxy(with_package: bool = False):
    """Oblivious Proxy Installation

    Args:
        with_package (bool, optional): Only available for .msi, .deb, .rpm. Defaults to False.
    """
    system_name = platform.system()
    if system_name == "Windows":
        windows_proxy_installation(with_package)
    elif system_name == "Linux":
        linux_proxy_installation(with_package)
    elif system_name == "Darwin":
        darwin_proxy_installation()


def windows_proxy_installation(with_package: bool = False):
    try:
        if with_package:
            url = "https://api.oblivious.ai/oblv-ccli/0.4.0/packages/oblv-0.4.0-x86_64.msi"
            res = requests.get(url)  # nosec
            path = os.path.join(os.path.expanduser("~"), "oblv-0.4.0-x86_64.msi")
            with open(path, "wb") as f:
                f.write(res.content)
            os.system("msiexec /I {} /quiet /QB-!".format(path))  # nosec
        else:
            url = "https://api.oblivious.ai/oblv-ccli/0.4.0/oblv-ccli-0.4.0-x86_64-pc-windows-msvc.zip"
            res = requests.get(url)  # nosec
            path = (
                os.getcwd().replace("\\", "/")
                + "/oblv-ccli-0.4.0-x86_64-pc-windows-msvc.zip"
            )
            with open(path, "wb") as f:
                f.write(res.content)
            with zipfile.ZipFile(path, "r") as zipObj:  # nosec
                zipObj.extractall()  # nosec
            os.environ["PATH"] += (
                ";" + os.getcwd() + "\\oblv-ccli-0.4.0-x86_64-pc-windows-msvc;"
            )
    except Exception as e:
        print(
            bcolors.RED
            + bcolors.BOLD
            + "Exception"
            + bcolors.BLACK
            + bcolors.ENDC
            + ": "
            + e.__cause__,
            file=sys.stderr,
        )


def linux_proxy_installation(with_package: bool = False):
    try:
        if with_package:
            try:
                os.system("dpkg")  # nosec
            except Exception:
                url = "https://api.oblivious.ai/oblv-ccli/0.4.0/packages/oblv-0.4.0-1.x86_64.rpm"
                res = requests.get(url)  # nosec
                path = os.path.join(os.path.expanduser("~"), "oblv-0.4.0-1.x86_64.rpm")
                with open(path, "wb") as f:
                    f.write(res.content)
                os.system("rpm -i {}".format(path))  # nosec
            else:
                url = "https://api.oblivious.ai/oblv-ccli/0.4.0/packages/oblv_0.4.0_amd64.deb"
                res = requests.get(url)  # nosec
                path = os.path.join(os.path.expanduser("~"), "oblv_0.4.0_amd64.deb")
                with open(path, "wb") as f:
                    f.write(res.content)
                os.system("dpkg -i {}".format(path))  # nosec
        else:
            url = "https://api.oblivious.ai/oblv-ccli/0.4.0/oblv-ccli-0.4.0-x86_64-unknown-linux-musl.tar.gz"
            file_name = "oblv-ccli-0.4.0-x86_64-unknown-linux-musl.tar.gz"
            res = requests.get(url, stream=True)  # nosec
            if res.status_code == 200:
                with open(file_name, "wb") as f:
                    f.write(res.raw.read())
            path = os.getcwd() + "/oblv-ccli-0.4.0-x86_64-unknown-linux-musl"
            file = tarfile.open(file_name)  # nosec
            file.extractall(path)  # nosec

            os.symlink(
                "/usr/local/bin/oblv",
                os.getcwd() + "/oblv-ccli-0.4.0-x86_64-unknown-linux-musl/oblv",
            )
            print(
                bcolors.green(bcolors.bold("Successfully")) + " installed Oblivous CLI"
            )
    except Exception as e:
        print(
            bcolors.RED
            + bcolors.BOLD
            + "Exception"
            + bcolors.BLACK
            + bcolors.ENDC
            + ": "
            + e.__cause__,
            file=sys.stderr,
        )


def darwin_proxy_installation():
    url = "https://api.oblivious.ai/oblv-ccli/0.4.0/oblv-ccli-0.4.0-x86_64-apple-darwin.tar.gz"
    file_name = "oblv-ccli-0.4.0-x86_64-apple-darwin.tar.gz"
    res = requests.get(url, stream=True)  # nosec
    if res.status_code == 200:
        with open(file_name, "wb") as f:
            f.write(res.raw.read())
    path = os.getcwd() + "/oblv-ccli-0.4.0-x86_64-apple-darwin"
    file = tarfile.open(file_name)
    file.extractall(path)  # nosec

    os.symlink(
        "/usr/local/bin/oblv", os.getcwd() + "/oblv-ccli-0.4.0-x86_64-apple-darwin/oblv"
    )
    print(bcolors.green(bcolors.bold("Successfully")) + " installed Oblivous CLI")


def create_oblv_key_pair(key_name):
    if check_oblv_proxy_installation_status() is None:
        return
    try:
        file_path = os.path.join(os.path.expanduser("~"), ".ssh", key_name)
        result = subprocess.run(  # nosec
            ["oblv", "keygen", "--key-name", key_name, "--output", file_path],
            capture_output=True,
        )
        if result.stderr:
            raise subprocess.CalledProcessError(  # nosec
                returncode=result.returncode, cmd=result.args, stderr=result.stderr
            )
        result = result.stdout.strip()
        return get_oblv_public_key(key_name)
    except Exception as e:
        raise Exception(e)


def get_oblv_public_key(key_name):
    try:
        filepath = os.path.join(
            os.path.expanduser("~"), ".ssh", key_name, key_name + "_public.der"
        )
        with open(filepath, "rb") as f:
            public_key = f.read()
        public_key = base64.encodebytes(public_key).decode("UTF-8").replace("\n", "")
        return public_key
    except FileNotFoundError:
        print(
            bcolors.RED
            + bcolors.BOLD
            + "Exception"
            + bcolors.BLACK
            + bcolors.ENDC
            + ": "
            + "No key found with given name",
            file=sys.stderr,
        )
        raise FileNotFoundError
    except Exception as e:
        raise Exception(e)



# future
from __future__ import annotations

# stdlib
from concurrent import futures
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

# third party
import pandas as pd
import requests

# relative
from .core.node.new.grid_url import GridURL
from .core.node.new.network_service import NodePeer
from .core.node.new.node_metadata import NodeMetadataJSON
from .logger import error
from .logger import warning

if TYPE_CHECKING:
    # relative
    from .core.node.common.client import Client

NETWORK_REGISTRY_URL = (
    "https://raw.githubusercontent.com/OpenMined/NetworkRegistry/main/gateways.json"
)
NETWORK_REGISTRY_REPO = "https://github.com/OpenMined/NetworkRegistry"


class NetworkRegistry:
    def __init__(self) -> None:
        self.all_networks: List[Dict] = []
        try:
            response = requests.get(NETWORK_REGISTRY_URL)  # nosec
            network_json = response.json()
            self.all_networks = network_json["2.0.0"]["gateways"]
        except Exception as e:
            warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. {e}"
            )

    @property
    def online_networks(self) -> List[Dict]:
        networks = self.all_networks

        def check_network(network: Dict) -> Optional[Dict[Any, Any]]:
            url = "http://" + network["host_or_ip"] + ":" + str(network["port"]) + "/"
            try:
                res = requests.get(url, timeout=0.5)  # nosec
                online = "This is a PyGrid Network node." in res.text
            except Exception:
                online = False

            # networks without frontend have a /ping route in 0.7.0
            if not online:
                try:
                    ping_url = url + "ping"
                    res = requests.get(ping_url, timeout=0.5)  # nosec
                    online = res.status_code == 200
                except Exception:
                    online = False

            if online:
                version = network.get("version", None)
                # Check if syft version was described in NetworkRegistry
                # If it's unknown, try to update it to an available version.
                if not version or version == "unknown":
                    # If not defined, try to ask in /syft/version endpoint (supported by 0.7.0)
                    try:
                        version_url = url + "api/v1/new/metadata"
                        res = requests.get(version_url, timeout=0.5)  # nosec
                        if res.status_code == 200:
                            network["version"] = res.json()["syft_version"]
                        else:
                            network["version"] = "unknown"
                    except Exception:
                        network["version"] = "unknown"
                return network
            return None

        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _online_networks = list(
                executor.map(lambda network: check_network(network), networks)
            )

        online_networks = list()
        for each in _online_networks:
            if each is not None:
                online_networks.append(each)
        return online_networks

    def _repr_html_(self) -> str:
        on = self.online_networks
        if len(on) == 0:
            return "(no gateways online - try syft.gateways.all_gateways to see offline gateways)"
        return pd.DataFrame(on)._repr_html_()

    def __repr__(self) -> str:
        on = self.online_networks
        if len(on) == 0:
            return "(no gateways online - try syft.gateways.all_gateways to see offline gateways)"
        return pd.DataFrame(on).to_string()

    @staticmethod
    def create_client(network: Dict[str, Any]) -> Client:  # type: ignore
        # relative
        from .core.node.new.client import connect

        try:
            port = int(network["port"])
            protocol = network["protocol"]
            host_or_ip = network["host_or_ip"]
            grid_url = GridURL(port=port, protocol=protocol, host_or_ip=host_or_ip)
            client = connect(url=str(grid_url))
            return client.guest()
        except Exception as e:
            error(f"Failed to login with: {network}. {e}")
            raise e

    def __getitem__(self, key: Union[str, int]) -> Client:  # type: ignore
        if isinstance(key, int):
            return self.create_client(network=self.online_networks[key])
        else:
            on = self.online_networks
            for network in on:
                if network["name"] == key:
                    return self.create_client(network=network)
        raise KeyError(f"Invalid key: {key} for {on}")


class DomainRegistry:
    def __init__(self) -> None:
        self.all_networks: List[Dict] = []
        self.all_domains: List = []
        try:
            response = requests.get(NETWORK_REGISTRY_URL)  # nosec
            network_json = response.json()
            self.all_networks = network_json["2.0.0"]["gateways"]
        except Exception as e:
            warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. {e}"
            )

    @property
    def online_networks(self) -> List[Dict]:
        networks = self.all_networks

        def check_network(network: Dict) -> Optional[Dict[Any, Any]]:
            url = "http://" + network["host_or_ip"] + ":" + str(network["port"]) + "/"
            try:
                res = requests.get(url, timeout=0.5)
                online = "This is a PyGrid Network node." in res.text
            except Exception:
                online = False

            # networks without frontend have a /ping route in 0.7.0
            if not online:
                try:
                    ping_url = url + "ping"
                    res = requests.get(ping_url, timeout=0.5)
                    online = res.status_code == 200
                except Exception:
                    online = False

            if online:
                version = network.get("version", None)
                # Check if syft version was described in NetworkRegistry
                # If it's unknown, try to update it to an available version.
                if not version or version == "unknown":
                    # If not defined, try to ask in /syft/version endpoint (supported by 0.7.0)
                    try:
                        version_url = url + "api/v1/new/metadata"
                        res = requests.get(version_url, timeout=0.5)
                        if res.status_code == 200:
                            network["version"] = res.json()["syft_version"]
                        else:
                            network["version"] = "unknown"
                    except Exception:
                        network["version"] = "unknown"
                return network
            return None

        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _online_networks = list(
                executor.map(lambda network: check_network(network), networks)
            )

        online_networks = list()
        for each in _online_networks:
            if each is not None:
                online_networks.append(each)
        return online_networks

    @property
    def online_domains(self) -> List[Tuple[NodePeer, NodeMetadataJSON]]:
        def check_domain(peer: NodePeer) -> Optional[Tuple[NodePeer, NodeMetadataJSON]]:
            try:
                guest_client = peer.guest_client
                metadata = guest_client.metadata
                return peer, metadata
            except Exception:  # nosec
                pass
            return None

        networks = self.online_networks

        self.all_domains = []
        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _all_online_domains = []
            for network in networks:
                network_client = NetworkRegistry.create_client(network)
                domains = network_client.domains
                self.all_domains += domains
                _online_domains = list(
                    executor.map(lambda domain: check_domain(domain), domains)
                )
                _all_online_domains += _online_domains

        online_domains = list()
        for each in _all_online_domains:
            if each is not None:
                online_domains.append(each)
        return online_domains

    def __make_dict__(self) -> List[Dict[str, str]]:
        on = self.online_domains
        domains = []
        domain_dict = {}
        for domain, metadata in on:
            domain_dict["name"] = domain.name
            domain_dict["organization"] = metadata.organization
            domain_dict["version"] = metadata.syft_version
            route = None
            if len(domain.node_routes) > 0:
                route = domain.node_routes[0]
            domain_dict["host_or_ip"] = route.host_or_ip if route else "-"
            domain_dict["protocol"] = route.protocol if route else "-"
            domain_dict["port"] = route.port if route else "-"
            domain_dict["id"] = domain.id
            domains.append(domain_dict)
        return domains

    def _repr_html_(self) -> str:
        on = self.__make_dict__()
        if len(on) == 0:
            return "(no domains online - try syft.domains.all_domains to see offline domains)"
        return pd.DataFrame(on)._repr_html_()

    def __repr__(self) -> str:
        on = self.__make_dict__()
        if len(on) == 0:
            return "(no domains online - try syft.domains.all_domains to see offline domains)"
        return pd.DataFrame(on).to_string()

    def create_client(self, peer: NodePeer) -> Client:  # type: ignore
        try:
            return peer.guest_client
        except Exception as e:
            error(f"Failed to login to: {peer}. {e}")
            raise e

    def __getitem__(self, key: Union[str, int]) -> Client:  # type: ignore
        if isinstance(key, int):
            return self.create_client(self.online_domains[key][0])
        else:
            on = self.online_domains
            count = 0
            for domain, _ in on:
                if domain.name == key:
                    return self.create_client(self.online_domains[count][0])
                count += 1
        raise KeyError(f"Invalid key: {key} for {on}")



# stdlib
from typing import List
from typing import Optional
from typing import Union

# third party
from result import Result

# relative
from ....telemetry import instrument
from .context import AuthedServiceContext
from .data_subject_member import ChildPartitionKey
from .data_subject_member import DataSubjectMemberRelationship
from .data_subject_member import ParentPartitionKey
from .document_store import BaseUIDStoreStash
from .document_store import DocumentStore
from .document_store import PartitionSettings
from .document_store import QueryKeys
from .response import SyftError
from .response import SyftSuccess
from .serializable import serializable
from .service import AbstractService
from .service import SERVICE_TO_TYPES
from .service import TYPE_TO_SERVICE


@instrument
@serializable(recursive_serde=True)
class DataSubjectMemberStash(BaseUIDStoreStash):
    object_type = DataSubjectMemberRelationship
    settings: PartitionSettings = PartitionSettings(
        name=DataSubjectMemberRelationship.__canonical_name__,
        object_type=DataSubjectMemberRelationship,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_all_for_parent(
        self, name: str
    ) -> Result[Optional[DataSubjectMemberRelationship], str]:
        qks = QueryKeys(qks=[ParentPartitionKey.with_obj(name)])
        return self.query_all(qks=qks)

    def get_all_for_child(
        self, name: str
    ) -> Result[Optional[DataSubjectMemberRelationship], str]:
        qks = QueryKeys(qks=[ChildPartitionKey.with_obj(name)])
        return self.query_all(qks=qks)


@instrument
@serializable(recursive_serde=True)
class DataSubjectMemberService(AbstractService):
    store: DocumentStore
    stash: DataSubjectMemberStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = DataSubjectMemberStash(store=store)

    def add(
        self, context: AuthedServiceContext, parent: str, child: str
    ) -> Union[SyftSuccess, SyftError]:
        """Register relationship between data subject and it's member."""
        relation = DataSubjectMemberRelationship(parent=parent, child=child)
        result = self.stash.set(relation, ignore_duplicates=True)
        if result.is_err():
            return SyftError(result.err())
        return SyftSuccess(message=f"Relationship added for: {parent} -> {child}")

    def get_relatives(
        self, context: AuthedServiceContext, data_subject_name: str
    ) -> Union[List[str], SyftError]:
        """Get all Members for given data subject"""
        result = self.stash.get_all_for_parent(name=data_subject_name)
        if result.is_ok():
            data_subject_members = result.ok()
            return data_subject_members
        return SyftError(message=result.err())


TYPE_TO_SERVICE[DataSubjectMemberRelationship] = DataSubjectMemberService
SERVICE_TO_TYPES[DataSubjectMemberService].update({DataSubjectMemberRelationship})



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

# third party
from pymongo import ASCENDING
from pymongo import WriteConcern
from pymongo.errors import DuplicateKeyError
from result import Err
from result import Ok
from result import Result

# relative
from .document_store import DocumentStore
from .document_store import QueryKey
from .document_store import QueryKeys
from .document_store import StoreClientConfig
from .document_store import StoreConfig
from .document_store import StorePartition
from .mongo_client import MongoClient
from .response import SyftSuccess
from .serializable import serializable
from .syft_object import StorableObjectType
from .syft_object import SyftObject
from .syft_object import SyftObjectRegistry
from .transforms import TransformContext
from .transforms import transform
from .transforms import transform_method


class MongoBsonObject(StorableObjectType, dict):
    pass


def _repr_debug_(value: Any) -> str:
    if hasattr(value, "_repr_debug_"):
        return value._repr_debug_()
    return repr(value)


def to_mongo(context: TransformContext) -> TransformContext:
    output = {}
    unique_keys_dict = context.obj._syft_unique_keys_dict()
    search_keys_dict = context.obj._syft_searchable_keys_dict()
    all_dict = unique_keys_dict
    all_dict.update(search_keys_dict)
    for k in all_dict:
        value = getattr(context.obj, k, "")
        # if the value is a method, store its value
        if callable(value):
            output[k] = value()
        else:
            output[k] = value

    output["_id"] = context.output["id"]
    output["__canonical_name__"] = context.obj.__canonical_name__
    output["__version__"] = context.obj.__version__
    output["__obj__"] = context.obj.to_dict()
    output["__arepr__"] = _repr_debug_(context.obj)  # a comes first in alphabet
    context.output = output
    return context


@transform(SyftObject, MongoBsonObject)
def syft_obj_to_mongo():
    return [to_mongo]


@transform_method(MongoBsonObject, SyftObject)
def from_mongo(
    storage_obj: Dict, context: Optional[TransformContext] = None
) -> SyftObject:
    constructor = SyftObjectRegistry.versioned_class(
        name=storage_obj["__canonical_name__"], version=storage_obj["__version__"]
    )
    if constructor is None:
        raise ValueError(
            "Versioned class should not be None for initialization of SyftObject."
        )
    output = storage_obj["__obj__"]
    for attr, funcs in constructor.__serde_overrides__.items():
        if attr in output:
            output[attr] = funcs[1](output[attr])
    return constructor(**output)


@serializable(recursive_serde=True)
class MongoStorePartition(StorePartition):
    __attr_allowlist__ = [
        "storage_type",
        "settings",
        "store_config",
        "unique_cks",
        "searchable_cks",
    ]
    storage_type: StorableObjectType = MongoBsonObject

    def init_store(self):
        super().init_store()
        self._init_collection()

    def _init_collection(self):
        client = MongoClient.from_config(config=self.store_config.client_config)
        self._collection = client.with_collection(
            collection_settings=self.settings, store_config=self.store_config
        )
        self._create_update_index()

    @property
    def collection(self):
        if not hasattr(self, "_collection"):
            self.init_store()

        return self._collection

    def _create_update_index(self):
        """Create or update mongo database indexes"""

        def check_index_keys(current_keys, new_index_keys):
            current_keys.sort()
            new_index_keys.sort()
            return current_keys == new_index_keys

        syft_obj = self.settings.object_type

        unique_attrs = getattr(syft_obj, "__attr_unique__", [])
        object_name = syft_obj.__canonical_name__

        new_index_keys = [(attr, ASCENDING) for attr in unique_attrs]

        current_indexes = self.collection.index_information()
        index_name = f"{object_name}_index_name"

        current_index_keys = current_indexes.get(index_name, None)

        if current_index_keys is not None:
            keys_same = check_index_keys(current_index_keys["key"], new_index_keys)
            if keys_same:
                return

            # Drop current index, since incompatible with current object
            try:
                self.collection.drop_index(index_or_name=index_name)
            except Exception as e:
                print(
                    f"Failed to drop index for object: {object_name} with index keys: {current_index_keys}"
                )
                raise e

        # If no new indexes, then skip index creation
        if len(new_index_keys) == 0:
            return

        try:
            self.collection.create_index(new_index_keys, unique=True, name=index_name)
            print(f"Index Created for {object_name} with indexes: {new_index_keys}")
        except Exception as e:
            print(
                f"Failed to create index for {object_name} with index keys: {new_index_keys}"
            )
            raise e

    def set(
        self,
        obj: SyftObject,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        storage_obj = obj.to(self.storage_type)

        collection = self.collection

        if ignore_duplicates:
            collection = collection.with_options(write_concern=WriteConcern(w=0))
        try:
            collection.insert_one(storage_obj)
        except DuplicateKeyError as e:
            return Err(f"Duplicate Key Error for {obj}: {e}")

        return Ok(obj)

    def update(self, qk: QueryKey, obj: SyftObject) -> Result[SyftObject, str]:
        storage_obj = obj.to(self.storage_type)
        try:
            result = self.collection.update_one(
                filter=qk.as_dict_mongo, update={"$set": storage_obj}
            )
        except Exception as e:
            return Err(f"Failed to update obj: {obj} with qk: {qk}. Error: {e}")

        if result.matched_count == 0:
            return Err(f"No object found with query key: {qk}")
        # elif result.modified_count == 0:
        #     return Err(f"Failed to modify obj: {obj} with qk: {qk}")
        return Ok(obj)

    def find_index_or_search_keys(
        self, index_qks: QueryKeys, search_qks: QueryKeys
    ) -> Result[List[SyftObject], str]:
        # TODO: pass index as hint to find method
        qks = QueryKeys(qks=(index_qks.all + search_qks.all))
        return self.get_all_from_store(qks=qks)

    def get_all_from_store(self, qks: QueryKeys) -> Result[List[SyftObject], str]:
        storage_objs = self.collection.find(filter=qks.as_dict_mongo)
        syft_objs = []
        for storage_obj in storage_objs:
            obj = self.storage_type(storage_obj)
            transform_context = TransformContext(output={}, obj=obj)
            syft_objs.append(obj.to(self.settings.object_type, transform_context))
        return Ok(syft_objs)

    def delete(self, qk: QueryKey) -> Result[SyftSuccess, Err]:
        qks = QueryKeys(qks=qk)
        result = self.collection.delete_one(filter=qks.as_dict_mongo)

        if result.deleted_count == 1:
            return Ok(SyftSuccess(message="Deleted"))

        return Err(f"Failed to delete object with qk: {qk}")

    def all(self):
        qks = QueryKeys(qks=())
        return self.get_all_from_store(qks=qks)

    def __len__(self):
        return self._collection.count_documents(filter={})


@serializable(recursive_serde=True)
class MongoDocumentStore(DocumentStore):
    partition_type = MongoStorePartition


@serializable(recursive_serde=True)
class MongoStoreConfig(StoreConfig):
    client_config: StoreClientConfig
    store_type: Type[DocumentStore] = MongoDocumentStore
    db_name: str = "app"



# future
from __future__ import annotations

# stdlib
from functools import partial
import types
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import _GenericAlias

# third party
from pydantic import BaseModel
from result import Err
from result import Ok
from result import Result
from typeguard import check_type

# relative
from ....telemetry import instrument
from .base import SyftBaseModel
from .response import SyftSuccess
from .serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftBaseObject
from .syft_object import SyftObject
from .uid import UID


@serializable(recursive_serde=True)
class BasePartitionSettings(SyftBaseModel):
    name: str


def first_or_none(result: Any) -> Optional[Any]:
    if hasattr(result, "__len__") and len(result) > 0:
        return Ok(result[0])
    return Ok(None)


class StoreClientConfig(BaseModel):
    pass


@serializable(recursive_serde=True)
class PartitionKey(BaseModel):
    key: str
    type_: Union[type, object]

    def __eq__(self, other: Any) -> bool:
        if type(other) == type(self):
            return self.key == other.key and self.type_ == other.type_
        return False

    def with_obj(self, obj: SyftObject) -> QueryKey:
        return QueryKey.from_obj(partition_key=self, obj=obj)

    def is_valid_list(self, obj: SyftObject) -> bool:
        # not a list and matches the internal list type of the _GenericAlias
        if not isinstance(obj, list):
            if not isinstance(obj, self.type_.__args__):
                obj = getattr(obj, self.key)
                if isinstance(obj, (types.FunctionType, types.MethodType)):
                    obj = obj()

            if not isinstance(obj, list) and isinstance(obj, self.type_.__args__):
                # still not a list but the right type
                obj = [obj]

        # is a list type so lets compare directly
        check_type("obj", obj, self.type_)
        return obj

    @property
    def type_list(self) -> bool:
        if isinstance(self.type_, _GenericAlias) and self.type_.__origin__ == list:
            return True
        return False


@serializable(recursive_serde=True)
class PartitionKeys(BaseModel):
    pks: Union[PartitionKey, Tuple[PartitionKey, ...]]

    @property
    def all(self) -> Iterable[PartitionKey]:
        # make sure we always return Tuple's even if theres a single value
        _keys = self.pks if isinstance(self.pks, (tuple, list)) else (self.pks,)
        return _keys

    def with_obj(self, obj: SyftObject) -> QueryKeys:
        return QueryKeys.from_obj(partition_keys=self, obj=obj)

    def with_tuple(self, *args: Tuple[Any, ...]) -> QueryKeys:
        return QueryKeys.from_tuple(partition_keys=self, args=args)

    def add(self, pk: PartitionKey) -> PartitionKeys:
        return PartitionKeys(pks=list(self.all) + [pk])

    @staticmethod
    def from_dict(cks_dict: Dict[str, type]) -> PartitionKeys:
        pks = []
        for k, t in cks_dict.items():
            pks.append(PartitionKey(key=k, type_=t))
        return PartitionKeys(pks=pks)

    def make(self, *obj_arg: Union[SyftObject, Tuple[Any, ...]]) -> QueryKeys:
        if isinstance(obj_arg, SyftObject):
            return self.with_obj(obj_arg)
        else:
            return self.with_tuple(*obj_arg)


@serializable(recursive_serde=True)
class QueryKey(PartitionKey):
    value: Any

    def __eq__(self, other: Any) -> bool:
        if type(other) == type(self):
            return (
                self.key == other.key
                and self.type_ == other.type_
                and self.value == other.value
            )
        return False

    @property
    def partition_key(self) -> PartitionKey:
        return PartitionKey(key=self.key, type_=self.type_)

    @staticmethod
    def from_obj(partition_key: PartitionKey, obj: SyftObject) -> List[Any]:
        pk_key = partition_key.key
        pk_type = partition_key.type_

        # 🟡 TODO: support more advanced types than List[type]
        if partition_key.type_list:
            pk_value = partition_key.is_valid_list(obj)
        else:
            if isinstance(obj, pk_type):
                pk_value = obj
            else:
                pk_value = getattr(obj, pk_key)
                # object has a method for getting these types
                # we can't use properties because we don't seem to be able to get the
                # return types
                if isinstance(pk_value, (types.FunctionType, types.MethodType)):
                    pk_value = pk_value()

            if pk_value and not isinstance(pk_value, pk_type):
                raise Exception(
                    f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                )
        return QueryKey(key=pk_key, type_=pk_type, value=pk_value)

    @property
    def as_dict(self):
        return {self.key: self.value}

    @property
    def as_dict_mongo(self):
        key = self.key
        if key == "id":
            key = "_id"
        if self.type_list:
            # We want to search inside the list of values
            return {key: {"$in": self.value}}
        return {key: self.value}


@serializable(recursive_serde=True)
class PartitionKeysWithUID(PartitionKeys):
    uid_pk: PartitionKey

    @property
    def all(self) -> Iterable[PartitionKey]:
        all_keys = self.pks if isinstance(self.pks, (tuple, list)) else [self.pks]
        if self.uid_pk not in all_keys:
            all_keys.insert(0, self.uid_pk)
        return all_keys


@serializable(recursive_serde=True)
class QueryKeys(SyftBaseModel):
    qks: Union[QueryKey, Tuple[QueryKey, ...]]

    @property
    def all(self) -> Iterable[QueryKey]:
        # make sure we always return Tuple's even if theres a single value
        _keys = self.qks if isinstance(self.qks, (tuple, list)) else (self.qks,)
        return _keys

    @staticmethod
    def from_obj(partition_keys: PartitionKeys, obj: SyftObject) -> QueryKeys:
        qks = []
        for partition_key in partition_keys.all:
            pk_key = partition_key.key
            pk_type = partition_key.type_
            pk_value = getattr(obj, pk_key)
            # object has a method for getting these types
            # we can't use properties because we don't seem to be able to get the
            # return types
            if isinstance(pk_value, (types.FunctionType, types.MethodType)):
                pk_value = pk_value()
            if partition_key.type_list:
                pk_value = partition_key.is_valid_list(obj)
            else:
                if pk_value and not isinstance(pk_value, pk_type):
                    raise Exception(
                        f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                    )
            qk = QueryKey(key=pk_key, type_=pk_type, value=pk_value)
            qks.append(qk)
        return QueryKeys(qks=qks)

    @staticmethod
    def from_tuple(partition_keys: PartitionKeys, args: Tuple[Any, ...]) -> QueryKeys:
        qks = []
        for partition_key, pk_value in zip(partition_keys.all, args):
            pk_key = partition_key.key
            pk_type = partition_key.type_
            if not isinstance(pk_value, pk_type):
                raise Exception(
                    f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                )
            qk = QueryKey(key=pk_key, type_=pk_type, value=pk_value)
            qks.append(qk)
        return QueryKeys(qks=qks)

    @staticmethod
    def from_dict(qks_dict: Dict[str, Any]) -> QueryKeys:
        qks = []
        for k, v in qks_dict.items():
            qks.append(QueryKey(key=k, type_=type(v), value=v))
        return QueryKeys(qks=qks)

    @property
    def as_dict(self):
        qk_dict = {}
        for qk in self.all:
            qk_key = qk.key
            qk_value = qk.value
            qk_dict[qk_key] = qk_value
        return qk_dict

    @property
    def as_dict_mongo(self):
        qk_dict = {}
        for qk in self.all:
            qk_key = qk.key
            qk_value = qk.value
            if qk_key == "id":
                qk_key = "_id"
            if qk.type_list:
                # We want to search inside the list of values
                qk_dict[qk_key] = {"$in": qk_value}
            else:
                qk_dict[qk_key] = qk_value
        return qk_dict


UIDPartitionKey = PartitionKey(key="id", type_=UID)


@serializable(recursive_serde=True)
class PartitionSettings(BasePartitionSettings):
    object_type: type
    store_key: PartitionKey = UIDPartitionKey

    @property
    def unique_keys(self) -> PartitionKeys:
        unique_keys = PartitionKeys.from_dict(self.object_type._syft_unique_keys_dict())
        return unique_keys.add(self.store_key)

    @property
    def searchable_keys(self) -> PartitionKeys:
        return PartitionKeys.from_dict(self.object_type._syft_searchable_keys_dict())


@instrument
@serializable(recursive_serde=True)
class StorePartition:
    def __init__(
        self,
        settings: PartitionSettings,
        store_config: StoreConfig,
    ) -> None:
        self.settings = settings
        self.store_config = store_config
        self.init_store()

    def init_store(self) -> None:
        self.unique_cks = self.settings.unique_keys.all
        self.searchable_cks = self.settings.searchable_keys.all

    def matches_unique_cks(self, partition_key: PartitionKey) -> bool:
        if partition_key in self.unique_cks:
            return True
        return False

    def matches_searchable_cks(self, partition_key: PartitionKey) -> bool:
        if partition_key in self.searchable_cks:
            return True
        return False

    def store_query_key(self, obj: Any) -> QueryKey:
        return self.settings.store_key.with_obj(obj)

    def store_query_keys(self, objs: Any) -> QueryKeys:
        return QueryKeys(qks=[self.store_query_key(obj) for obj in objs])

    def find_index_or_search_keys(self, index_qks: QueryKeys, search_qks: QueryKeys):
        raise NotImplementedError

    def all(self) -> Result[List[BaseStash.object_type], str]:
        raise NotImplementedError

    def set(
        self,
        obj: SyftObject,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        raise NotImplementedError

    def update(self, qk: QueryKey, obj: SyftObject) -> Result[SyftObject, str]:
        raise NotImplementedError

    def get_all_from_store(self, qks: QueryKeys) -> Result[List[SyftObject], str]:
        raise NotImplementedError

    def create(self, obj: SyftObject) -> Result[SyftObject, str]:
        raise NotImplementedError

    def delete(self, qk: QueryKey) -> Result[SyftSuccess, Err]:
        raise NotImplementedError


@instrument
@serializable(recursive_serde=True)
class DocumentStore:
    partitions: Dict[str, StorePartition]
    partition_type: Type[StorePartition]

    def __init__(self, store_config: StoreConfig) -> None:
        if store_config is None:
            raise Exception("must have store config")
        self.partitions = {}
        self.store_config = store_config

    def partition(self, settings: PartitionSettings) -> StorePartition:
        if settings.name not in self.partitions:
            self.partitions[settings.name] = self.partition_type(
                settings=settings, store_config=self.store_config
            )
        return self.partitions[settings.name]


@instrument
class BaseStash:
    object_type: Type[SyftObject]
    settings: PartitionSettings
    partition: StorePartition

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.partition = store.partition(type(self).settings)

    def check_type(self, obj: Any, type_: type) -> Result[Any, str]:
        return (
            Ok(obj)
            if isinstance(obj, type_)
            else Err(f"{type(obj)} does not match required type: {type_}")
        )

    def get_all(self) -> Result[List[BaseStash.object_type], str]:
        return self.partition.all()

    def __len__(self) -> int:
        return self.partition.__len__()

    def set(
        self,
        obj: BaseStash.object_type,
        ignore_duplicates: bool = False,
    ) -> Result[BaseStash.object_type, str]:
        return self.partition.set(obj=obj, ignore_duplicates=ignore_duplicates)

    def query_all(
        self, qks: Union[QueryKey, QueryKeys]
    ) -> Result[List[BaseStash.object_type], str]:
        if isinstance(qks, QueryKey):
            qks = QueryKeys(qks=qks)

        unique_keys = []
        searchable_keys = []

        for qk in qks.all:
            pk = qk.partition_key
            if self.partition.matches_unique_cks(pk):
                unique_keys.append(qk)
            elif self.partition.matches_searchable_cks(pk):
                searchable_keys.append(qk)
            else:
                return Err(
                    f"{qk} not in {type(self.partition)} unique or searchable keys"
                )

        index_qks = QueryKeys(qks=unique_keys)
        search_qks = QueryKeys(qks=searchable_keys)
        return self.partition.find_index_or_search_keys(
            index_qks=index_qks, search_qks=search_qks
        )

    def query_all_kwargs(
        self, **kwargs: Dict[str, Any]
    ) -> Result[List[BaseStash.object_type], str]:
        qks = QueryKeys.from_dict(kwargs)
        return self.query_all(qks=qks)

    def query_one(
        self, qks: Union[QueryKey, QueryKeys]
    ) -> Result[Optional[BaseStash.object_type], str]:
        return self.query_all(qks=qks).and_then(first_or_none)

    def query_one_kwargs(
        self,
        **kwargs: Dict[str, Any],
    ) -> Result[Optional[BaseStash.object_type], str]:
        return self.query_all_kwargs(**kwargs).and_then(first_or_none)

    def find_all(
        self, **kwargs: Dict[str, Any]
    ) -> Result[List[BaseStash.object_type], str]:
        return self.query_all_kwargs(**kwargs)

    def find_one(
        self, **kwargs: Dict[str, Any]
    ) -> Result[Optional[BaseStash.object_type], str]:
        return self.query_one_kwargs(**kwargs)

    def find_and_delete(self, **kwargs: Dict[str, Any]) -> Result[SyftSuccess, Err]:
        obj = self.query_one_kwargs(**kwargs)
        if obj.is_err():
            return obj.err()
        else:
            obj = obj.ok()

        if not obj:
            return Err(f"Object does not exists with kwargs: {kwargs}")
        qk = self.partition.store_query_key(obj)
        return self.delete(qk=qk)

    def delete(self, qk: QueryKey) -> Result[SyftSuccess, Err]:
        return self.partition.delete(qk=qk)

    def update(
        self, obj: BaseStash.object_type
    ) -> Optional[Result[BaseStash.object_type, str]]:
        qk = self.partition.store_query_key(obj)
        return self.partition.update(qk=qk, obj=obj)


@instrument
class BaseUIDStoreStash(BaseStash):
    def delete_by_uid(self, uid: UID) -> Result[SyftSuccess, str]:
        qk = UIDPartitionKey.with_obj(uid)
        result = super().delete(qk=qk)
        if result.is_ok():
            return Ok(SyftSuccess(message=f"ID: {uid} deleted"))
        return result.err()

    def get_by_uid(
        self, uid: UID
    ) -> Result[Optional[BaseUIDStoreStash.object_type], str]:
        qks = QueryKeys(qks=[UIDPartitionKey.with_obj(uid)])
        return self.query_one(qks=qks)

    def set(
        self,
        obj: BaseUIDStoreStash.object_type,
        ignore_duplicates: bool = False,
    ) -> Result[BaseUIDStoreStash.object_type, str]:
        set_method = partial(super().set, ignore_duplicates=ignore_duplicates)
        return self.check_type(obj, self.object_type).and_then(set_method)


@serializable(recursive_serde=True)
class StoreConfig(SyftBaseObject):
    __canonical_name__ = "StoreConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    store_type: Type[DocumentStore]
    client_config: Optional[StoreClientConfig]



# stdlib
from enum import Enum
from typing import Optional

# relative
from .credentials import SyftVerifyKey
from .linked_obj import LinkedObject
from .request import DateTime
from .serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .transforms import TransformContext
from .transforms import add_credentials_for_key
from .transforms import add_node_uid_for_key
from .transforms import generate_id
from .transforms import transform
from .uid import UID


@serializable(recursive_serde=True)
class MessageStatus(Enum):
    UNDELIVERED = 0
    DELIVERED = 1


class MessageExpiryStatus(Enum):
    AUTO = 0
    NEVER = 1


@serializable(recursive_serde=True)
class Message(SyftObject):
    __canonical_name__ = "Message"
    __version__ = SYFT_OBJECT_VERSION_1

    subject: str
    node_uid: UID
    from_user_verify_key: SyftVerifyKey
    to_user_verify_key: SyftVerifyKey
    created_at: DateTime
    status: MessageStatus = MessageStatus.UNDELIVERED
    linked_obj: Optional[LinkedObject]

    __attr_searchable__ = [
        "from_user_verify_key",
        "to_user_verify_key",
        "status",
    ]
    __attr_repr_cols__ = ["subject", "status", "created_at", "linked_obj"]

    @property
    def link(self) -> Optional[SyftObject]:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None


@serializable(recursive_serde=True)
class CreateMessage(Message):
    __canonical_name__ = "CreateMessage"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    node_uid: Optional[UID]
    from_user_verify_key: Optional[SyftVerifyKey]
    created_at: Optional[DateTime]


def add_msg_creation_time(context: TransformContext) -> TransformContext:
    context.output["created_at"] = DateTime.now()
    return context


@transform(CreateMessage, Message)
def createmessage_to_message():
    return [
        generate_id,
        add_msg_creation_time,
        add_credentials_for_key("from_user_verify_key"),
        add_node_uid_for_key("node_uid"),
    ]



# future
from __future__ import annotations

# stdlib
import contextlib
from datetime import datetime
from functools import partial
import hashlib
import os
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
import gevent
import gipc
from gipc.gipc import _GIPCDuplexHandle
from nacl.signing import SigningKey
from result import Err
from result import Result

# relative
from ... import __version__
from ...external import OBLV
from ...telemetry import instrument
from ...util import random_name
from .new.action_service import ActionService
from .new.action_store import DictActionStore
from .new.action_store import SQLiteActionStore
from .new.api import SignedSyftAPICall
from .new.api import SyftAPI
from .new.api import SyftAPICall
from .new.api import SyftAPIData
from .new.context import AuthedServiceContext
from .new.context import NodeServiceContext
from .new.context import UnauthedServiceContext
from .new.context import UserLoginCredentials
from .new.credentials import SyftSigningKey
from .new.credentials import SyftVerifyKey
from .new.data_subject_member_service import DataSubjectMemberService
from .new.data_subject_service import DataSubjectService
from .new.dataset_service import DatasetService
from .new.deserialize import _deserialize
from .new.dict_document_store import DictStoreConfig
from .new.document_store import StoreConfig
from .new.message_service import MessageService
from .new.metadata_service import MetadataService
from .new.metadata_stash import MetadataStash
from .new.network_service import NetworkService
from .new.node import NewNode
from .new.node import NodeType
from .new.node_metadata import NodeMetadata
from .new.project_service import ProjectService
from .new.queue_stash import QueueItem
from .new.queue_stash import QueueStash
from .new.request_service import RequestService
from .new.response import SyftError
from .new.serializable import serializable
from .new.serialize import _serialize
from .new.service import AbstractService
from .new.service import ServiceConfigRegistry
from .new.sqlite_document_store import SQLiteStoreClientConfig
from .new.sqlite_document_store import SQLiteStoreConfig
from .new.syft_object import HIGHEST_SYFT_OBJECT_VERSION
from .new.syft_object import LOWEST_SYFT_OBJECT_VERSION
from .new.syft_object import SyftObject
from .new.test_service import TestService
from .new.uid import UID
from .new.user import ServiceRole
from .new.user import User
from .new.user import UserCreate
from .new.user_code_service import UserCodeService
from .new.user_service import UserService
from .new.user_stash import UserStash
from .new.worker_settings import WorkerSettings


def gipc_encoder(obj):
    return _serialize(obj, to_bytes=True)


def gipc_decoder(obj_bytes):
    return _deserialize(obj_bytes, from_bytes=True)


NODE_PRIVATE_KEY = "NODE_PRIVATE_KEY"
NODE_UID = "NODE_UID"


def get_private_key_env() -> Optional[str]:
    return get_env(NODE_PRIVATE_KEY)


def get_node_uid_env() -> Optional[str]:
    return get_env(NODE_UID)


def get_env(key: str) -> Optional[str]:
    return os.environ.get(key, None)


signing_key_env = get_private_key_env()
node_uid_env = get_node_uid_env()


@instrument
@serializable(recursive_serde=True)
class Worker(NewNode):
    signing_key: Optional[SyftSigningKey]
    required_signed_calls: bool = True

    def __init__(
        self,
        *,  # Trasterisk
        name: Optional[str] = None,
        id: Optional[UID] = None,
        services: Optional[List[Type[AbstractService]]] = None,
        signing_key: Optional[Union[SyftSigningKey, SigningKey]] = None,
        action_store_config: Optional[StoreConfig] = None,
        document_store_config: Optional[StoreConfig] = None,
        root_email: str = "info@openmined.org",
        root_password: str = "changethis",
        processes: int = 0,
        is_subprocess: bool = False,
        node_type: NodeType = NodeType.DOMAIN,
        local_db: bool = False,
        sqlite_path: Optional[str] = None,
    ):
        # 🟡 TODO 22: change our ENV variable format and default init args to make this
        # less horrible or add some convenience functions
        if node_uid_env is not None:
            self.id = UID.from_string(node_uid_env)
        else:
            if id is None:
                id = UID()
            self.id = id

        self.signing_key = None
        if signing_key_env is not None:
            self.signing_key = SyftSigningKey.from_string(signing_key_env)
        else:
            if isinstance(signing_key, SigningKey):
                signing_key = SyftSigningKey(signing_key=signing_key)
            self.signing_key = signing_key

        if self.signing_key is None:
            self.signing_key = SyftSigningKey.generate()

        self.processes = processes
        self.is_subprocess = is_subprocess
        if name is None:
            name = random_name()
        self.name = name
        services = (
            [
                UserService,
                MetadataService,
                ActionService,
                TestService,
                DatasetService,
                UserCodeService,
                RequestService,
                DataSubjectService,
                NetworkService,
                MessageService,
                ProjectService,
                DataSubjectMemberService,
            ]
            if services is None
            else services
        )

        self.service_config = ServiceConfigRegistry.get_registered_configs()
        self.local_db = local_db
        self.sqlite_path = sqlite_path
        self.init_stores(
            action_store_config=action_store_config,
            document_store_config=document_store_config,
        )

        if OBLV:
            # relative
            from ...external.oblv.oblv_service import OblvService

            services += [OblvService]
            create_oblv_key_pair(worker=self)

        self.services = services
        self._construct_services()

        create_admin_new(  # nosec B106
            name="Jane Doe",
            email="info@openmined.org",
            password="changethis",
            node=self,
        )

        self.client_cache = {}
        self.node_type = node_type

        self.post_init()

    @staticmethod
    def named(
        name: str,
        processes: int = 0,
        reset: bool = False,
        local_db: bool = False,
        sqlite_path: Optional[str] = None,
    ) -> Worker:
        name_hash = hashlib.sha256(name.encode("utf8")).digest()
        uid = UID(name_hash[0:16])
        key = SyftSigningKey(SigningKey(name_hash))
        if reset:
            store_config = SQLiteStoreClientConfig()
            store_config.filename = f"{uid}.sqlite"
            with contextlib.suppress(FileNotFoundError):
                if os.path.exists(store_config.file_path):
                    os.unlink(store_config.file_path)

        return Worker(
            name=name,
            id=uid,
            signing_key=key,
            processes=processes,
            local_db=local_db,
            sqlite_path=sqlite_path,
        )

    def is_root(self, credentials: SyftVerifyKey) -> bool:
        return credentials == self.signing_key.verify_key

    @property
    def root_client(self) -> Any:
        # relative
        from .new.client import PythonConnection
        from .new.client import SyftClient

        connection = PythonConnection(node=self)
        return SyftClient(connection=connection, credentials=self.signing_key)

    @property
    def guest_client(self) -> Any:
        # relative
        from .new.client import PythonConnection
        from .new.client import SyftClient

        connection = PythonConnection(node=self)
        return SyftClient(connection=connection, credentials=SyftSigningKey.generate())

    def __repr__(self) -> str:
        return f"{type(self).__name__}: {self.name} - {self.id} - {self.node_type} - {self.services}"

    def post_init(self) -> None:
        if self.is_subprocess:
            # print(f"> Starting Subprocess {self}")
            pass
        else:
            print(f"> Starting {self}")
        # super().post_init()

    def init_stores(
        self,
        document_store_config: Optional[StoreConfig] = None,
        action_store_config: Optional[StoreConfig] = None,
    ):
        if document_store_config is None:
            if self.local_db or (self.processes > 0 and not self.is_subprocess):
                client_config = SQLiteStoreClientConfig(path=self.sqlite_path)
                document_store_config = SQLiteStoreConfig(client_config=client_config)
            else:
                document_store_config = DictStoreConfig()
        if (
            isinstance(document_store_config, SQLiteStoreConfig)
            and document_store_config.client_config.filename is None
        ):
            document_store_config.client_config.filename
            document_store_config.client_config.filename = f"{self.id}.sqlite"
            print(
                f"SQLite Store Path:\n!open file://{document_store_config.client_config.file_path}\n"
            )
        document_store = document_store_config.store_type
        self.document_store_config = document_store_config

        self.document_store = document_store(store_config=document_store_config)
        if action_store_config is None:
            if self.local_db or (self.processes > 0 and not self.is_subprocess):
                client_config = SQLiteStoreClientConfig(path=self.sqlite_path)
                action_store_config = SQLiteStoreConfig(client_config=client_config)
            else:
                action_store_config = DictStoreConfig()

        if (
            isinstance(action_store_config, SQLiteStoreConfig)
            and action_store_config.client_config.filename is None
        ):
            action_store_config.client_config.filename = f"{self.id}.sqlite"

        if isinstance(action_store_config, SQLiteStoreConfig):
            self.action_store = SQLiteActionStore(
                store_config=action_store_config,
                root_verify_key=self.signing_key.verify_key,
            )
        else:
            self.action_store = DictActionStore(
                root_verify_key=self.signing_key.verify_key
            )

        self.action_store_config = action_store_config
        self.queue_stash = QueueStash(store=self.document_store)

    def _construct_services(self):
        self.service_path_map = {}

        for service_klass in self.services:
            kwargs = {}
            if service_klass == ActionService:
                kwargs["store"] = self.action_store
            store_services = [
                UserService,
                MetadataService,
                DatasetService,
                UserCodeService,
                RequestService,
                DataSubjectService,
                NetworkService,
                MessageService,
                ProjectService,
                DataSubjectMemberService,
            ]

            if OBLV:
                # relative
                from ...external.oblv.oblv_service import OblvService

                store_services += [OblvService]

            if service_klass in store_services:
                kwargs["store"] = self.document_store
            self.service_path_map[service_klass.__name__.lower()] = service_klass(
                **kwargs
            )

    def get_service_method(self, path_or_func: Union[str, Callable]) -> Callable:
        if callable(path_or_func):
            path_or_func = path_or_func.__qualname__
        return self._get_service_method_from_path(path_or_func)

    def get_service(self, path_or_func: Union[str, Callable]) -> Callable:
        if callable(path_or_func):
            path_or_func = path_or_func.__qualname__
        return self._get_service_from_path(path_or_func)

    def _get_service_from_path(self, path: str) -> AbstractService:
        path_list = path.split(".")
        if len(path_list) > 1:
            _ = path_list.pop()
        service_name = path_list.pop()
        return self.service_path_map[service_name.lower()]

    def _get_service_method_from_path(self, path: str) -> Callable:
        path_list = path.split(".")
        method_name = path_list.pop()
        service_obj = self._get_service_from_path(path=path)

        return getattr(service_obj, method_name)

    @property
    def metadata(self) -> NodeMetadata:
        return NodeMetadata(
            name=self.name,
            id=self.id,
            verify_key=self.signing_key.verify_key,
            highest_object_version=HIGHEST_SYFT_OBJECT_VERSION,
            lowest_object_version=LOWEST_SYFT_OBJECT_VERSION,
            syft_version=__version__,
        )

    @property
    def icon(self) -> str:
        return "🦾"

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Worker):
            return False

        if self.id != other.id:
            return False

        return True

    def resolve_future(self, uid: UID) -> Union[Optional[QueueItem], SyftError]:
        result = self.queue_stash.pop(uid)
        if result.is_ok():
            return result.ok()
        return result.err()

    def forward_message(
        self, api_call: Union[SyftAPICall, SignedSyftAPICall]
    ) -> Result[Union[QueueItem, SyftObject], Err]:
        node_uid = api_call.message.node_uid
        if NetworkService not in self.services:
            return SyftError(
                message=(
                    "Node has no network service so we can't "
                    f"forward this message to {node_uid}"
                )
            )

        client = None
        if node_uid in self.client_cache:
            client = self.client_cache[node_uid]
        else:
            network_service = self.get_service(NetworkService)
            peer = network_service.stash.get_by_uid(node_uid)

            if peer.is_ok() and peer.ok():
                peer = peer.ok()
                context = NodeServiceContext(node=self)
                client = peer.client_with_context(context=context)
                self.client_cache[node_uid] = client

        if client:
            return client.connection.make_call(api_call)

        return SyftError(message=(f"Node has no route to {node_uid}"))

    def handle_api_call(
        self, api_call: Union[SyftAPICall, SignedSyftAPICall]
    ) -> Result[SignedSyftAPICall, Err]:
        # Get the result
        result = self.handle_api_call_with_unsigned_result(api_call)
        # Sign the result
        signed_result = SyftAPIData(data=result).sign(self.signing_key)

        return signed_result

    def handle_api_call_with_unsigned_result(
        self, api_call: Union[SyftAPICall, SignedSyftAPICall]
    ) -> Result[Union[QueueItem, SyftObject], Err]:
        if self.required_signed_calls and isinstance(api_call, SyftAPICall):
            return SyftError(
                message=f"You sent a {type(api_call)}. This node requires SignedSyftAPICall."  # type: ignore
            )
        else:
            if not api_call.is_valid:
                return SyftError(message="Your message signature is invalid")  # type: ignore

        if api_call.message.node_uid != self.id:
            return self.forward_message(api_call=api_call)

        if api_call.message.path == "queue":
            return self.resolve_future(uid=api_call.message.kwargs["uid"])

        if api_call.message.path == "metadata":
            return self.metadata

        result = None
        if self.is_subprocess or self.processes == 0:
            credentials = api_call.credentials
            api_call = api_call.message

            context = AuthedServiceContext(node=self, credentials=credentials)

            # 🔵 TODO 4: Add @service decorator to autobind services into the SyftAPI
            if api_call.path not in self.service_config:
                return SyftError(message=f"API call not in registered services: {api_call.path}")  # type: ignore

            _private_api_path = ServiceConfigRegistry.private_path_for(api_call.path)
            method = self.get_service_method(_private_api_path)
            try:
                result = method(context, *api_call.args, **api_call.kwargs)
            except Exception as e:
                result = SyftError(message=f"Exception calling {api_call.path}. {e}")
        else:
            worker_settings = WorkerSettings(
                id=self.id,
                name=self.name,
                signing_key=self.signing_key,
                document_store_config=self.document_store_config,
                action_store_config=self.action_store_config,
            )

            task_uid = UID()
            item = QueueItem(id=task_uid, node_uid=self.id)
            # 🟡 TODO 36: Needs distributed lock
            # self.queue_stash.set_placeholder(item)
            # self.queue_stash.partition.commit()
            thread = gevent.spawn(
                queue_task,
                api_call,
                worker_settings,
                task_uid,
                api_call.message.blocking,
            )
            if api_call.message.blocking:
                gevent.joinall([thread])
                signed_result = thread.value

                if not signed_result.is_valid:
                    return SyftError(message="The result signature is invalid")  # type: ignore

                result = signed_result.message.data
            else:
                result = item
        return result

    def get_api(self) -> SyftAPI:
        return SyftAPI.for_user(node=self)

    def get_method_with_context(
        self, function: Callable, context: NodeServiceContext
    ) -> Callable:
        method = self.get_service_method(function)
        return partial(method, context=context)

    def get_unauthed_context(
        self, login_credentials: UserLoginCredentials
    ) -> NodeServiceContext:
        return UnauthedServiceContext(node=self, login_credentials=login_credentials)


def task_producer(
    pipe: _GIPCDuplexHandle, api_call: SyftAPICall, blocking: bool
) -> Any:
    try:
        result = None
        with pipe:
            pipe.put(api_call)
            gevent.sleep(0)
            if blocking:
                try:
                    result = pipe.get()
                except EOFError:
                    pass
            pipe.close()
        if blocking:
            return result
    except gipc.gipc.GIPCClosed:
        pass
    except Exception as e:
        print("Exception in task_producer", e)


def task_runner(
    pipe: _GIPCDuplexHandle,
    worker_settings: WorkerSettings,
    task_uid: UID,
    blocking: bool,
) -> None:
    worker = Worker(
        id=worker_settings.id,
        name=worker_settings.name,
        signing_key=worker_settings.signing_key,
        document_store_config=worker_settings.document_store_config,
        action_store_config=worker_settings.action_store_config,
        is_subprocess=True,
    )
    try:
        with pipe:
            api_call = pipe.get()

            result = worker.handle_api_call(api_call)
            if blocking:
                pipe.put(result)
            else:
                item = QueueItem(
                    node_uid=worker.id, id=task_uid, result=result, resolved=True
                )
                worker.queue_stash.set_result(item)
                worker.queue_stash.partition.close()
            pipe.close()
    except Exception as e:
        print("Exception in task_runner", e)
        raise e


def queue_task(
    api_call: SyftAPICall,
    worker_settings: WorkerSettings,
    task_uid: UID,
    blocking: bool,
) -> Optional[Any]:
    with gipc.pipe(encoder=gipc_encoder, decoder=gipc_decoder, duplex=True) as (
        cend,
        pend,
    ):
        process = gipc.start_process(
            task_runner, args=(cend, worker_settings, task_uid, blocking)
        )
        producer = gevent.spawn(task_producer, pend, api_call, blocking)
        try:
            process.join()
        except KeyboardInterrupt:
            producer.kill(block=True)
            process.terminate()
        process.join()

    if blocking:
        return producer.value
    return None


def create_worker_metadata(
    worker: NewNode,
) -> Optional[NodeMetadata]:
    try:
        metadata_stash = MetadataStash(store=worker.document_store)
        metadata_exists = metadata_stash.get_all().ok()
        if metadata_exists:
            return None
        else:
            new_metadata = NodeMetadata(
                name=worker.name,
                id=worker.id,
                verify_key=worker.signing_key.verify_key,
                highest_object_version=HIGHEST_SYFT_OBJECT_VERSION,
                lowest_object_version=LOWEST_SYFT_OBJECT_VERSION,
                syft_version=__version__,
                deployed_on=datetime.now().date().strftime("%m/%d/%Y"),
            )
            result = metadata_stash.set(metadata=new_metadata)
            if result.is_ok():
                return result.ok()
            return None
    except Exception as e:
        print("create_worker_metadata failed", e)


def create_admin_new(
    name: str,
    email: str,
    password: str,
    node: NewNode,
) -> Optional[User]:
    try:
        user_stash = UserStash(store=node.document_store)
        row_exists = user_stash.get_by_email(email=email).ok()
        if row_exists:
            return None
        else:
            create_user = UserCreate(
                name=name,
                email=email,
                password=password,
                password_verify=password,
                role=ServiceRole.ADMIN,
            )
            # New User Initialization
            # 🟡 TODO: change later but for now this gives the main user super user automatically
            user = create_user.to(User)
            user.signing_key = node.signing_key
            user.verify_key = user.signing_key.verify_key
            result = user_stash.set(user=user)
            if result.is_ok():
                return result.ok()
            return None
    except Exception as e:
        print("Unable to create new admin", e)


def create_oblv_key_pair(
    worker: Worker,
) -> Optional[str]:
    try:
        # relative
        from ...external.oblv.oblv_keys_stash import OblvKeys
        from ...external.oblv.oblv_keys_stash import OblvKeysStash
        from ...external.oblv.oblv_service import generate_oblv_key

        oblv_keys_stash = OblvKeysStash(store=worker.document_store)

        if not len(oblv_keys_stash):
            public_key, private_key = generate_oblv_key()
            oblv_keys = OblvKeys(public_key=public_key, private_key=private_key)
            res = oblv_keys_stash.set(oblv_keys)
            if res.is_ok():
                print("Successfully generated Oblv Key pair at startup")
            return res.err()
        else:
            print(f"Using Existing Public/Private Key pair: {len(oblv_keys_stash)}")
    except Exception as e:
        print("Unable to create Oblv Keys.", e)



# stdlib
from enum import Enum
from typing import Callable
from typing import List
from typing import Optional
from typing import Type

# third party
from result import OkErr

# relative
from .credentials import SyftVerifyKey
from .linked_obj import LinkedObject
from .request import EnumMutation
from .request import Request
from .request import SubmitRequest
from .request import UserCodeStatusChange
from .request_service import RequestService
from .response import SyftError
from .serializable import serializable
from .service import TYPE_TO_SERVICE
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .transforms import TransformContext
from .transforms import generate_id
from .transforms import transform
from .uid import UID
from .user_code import UserCode


@serializable(recursive_serde=True)
class Project(SyftObject):
    __canonical_name__ = "Project"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str
    description: str
    user_verify_key: SyftVerifyKey
    requests: Optional[List[Request]]

    __attr_searchable__ = [
        "user_verify_key",
        "name",
    ]
    __attr_unique__ = ["name"]

    __attr_repr_cols__ = ["requests"]


@serializable(recursive_serde=True)
class ObjectPermissionChange(SyftObject):
    __canonical_name__ = "PermissionChange"
    __version__ = SYFT_OBJECT_VERSION_1

    object_uid: UID
    permission: Enum
    object_type: Type[SyftObject]


@serializable(recursive_serde=True)
class ProjectSubmit(SyftObject):
    __canonical_name__ = "ProjectSubmit"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    name: str
    description: Optional[str]
    changes: List[ObjectPermissionChange] = []

    __attr_repr_cols__ = ["name", "changes"]

    def set_description(self, description: str, msg: Optional[str] = None) -> None:
        self.description = description

    def add_request(self, obj: SyftObject, permission: Enum) -> None:
        change = ObjectPermissionChange(
            object_uid=obj.id, object_type=type(obj), permission=permission
        )
        self.changes.append(change)


def submit_changes(context: TransformContext) -> TransformContext:
    changes = context.output.pop("changes", [])
    if changes:
        mutations = []
        for change in changes:
            service_type = TYPE_TO_SERVICE[change.object_type]
            linked_obj = LinkedObject(
                node_uid=context.node.id,
                service_type=service_type,
                object_type=change.object_type,
                object_uid=change.object_uid,
            )

            if change.object_type == UserCode:
                mutation = UserCodeStatusChange(
                    value=change.permission, linked_obj=linked_obj
                )
            else:
                mutation = EnumMutation.from_obj(
                    linked_obj=linked_obj, attr_name="status", value=change.permission
                )

            mutations.append(mutation)
        submit_request = SubmitRequest(changes=mutations)
        request_submit_method = context.node.get_service_method(RequestService.submit)
        result = request_submit_method(
            context=context, request=submit_request, send_message=False
        )

        if isinstance(result, SyftError):
            return result
        if isinstance(result, OkErr):
            result = result.ok()
        context.output["requests"] = [result]
    context.output["user_verify_key"] = context.credentials
    return context


@transform(ProjectSubmit, Project)
def projectsubmit_to_project() -> List[Callable]:
    return [generate_id, submit_changes]



# stdlib
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Union

# third party
import numpy as np
from result import Err
from result import Ok
from result import Result

# relative
from .action_object import Action
from .action_object import ActionObject
from .action_object import ActionObjectPointer
from .action_object import AnyActionObject
from .action_store import ActionStore
from .action_types import action_type_for_type
from .context import AuthedServiceContext
from .numpy import NumpyArrayObject
from .pandas import PandasDataFrameObject  # noqa: F401
from .pandas import PandasSeriesObject  # noqa: F401
from .response import SyftError
from .response import SyftSuccess
from .serializable import serializable
from .service import AbstractService
from .service import SERVICE_TO_TYPES
from .service import TYPE_TO_SERVICE
from .service import service_method
from .twin_object import TwinObject
from .uid import UID
from .user_code import UserCode
from .user_code import execute_byte_code


@serializable(recursive_serde=True)
class TwinMode(Enum):
    NONE = 0
    PRIVATE = 1
    MOCK = 2


@serializable(recursive_serde=True)
class ActionService(AbstractService):
    def __init__(self, store: ActionStore) -> None:
        self.store = store

    @service_method(path="action.np_array", name="np_array")
    def np_array(self, context: AuthedServiceContext, data: Any) -> Any:
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        np_obj = NumpyArrayObject(
            syft_action_data=data, dtype=data.dtype, shape=data.shape
        )
        np_pointer = self.set(context, np_obj)
        return np_pointer

    @service_method(path="action.set", name="set")
    def set(
        self,
        context: AuthedServiceContext,
        action_object: Union[ActionObject, TwinObject],
    ) -> Result[ActionObject, str]:
        """Save an object to the action store"""
        # 🟡 TODO 9: Create some kind of type checking / protocol for SyftSerializable
        result = self.store.set(
            uid=action_object.id,
            credentials=context.credentials,
            syft_object=action_object,
        )
        if result.is_ok():
            if isinstance(action_object, TwinObject):
                action_object = action_object.mock
            action_object.syft_point_to(context.node.id)
            return Ok(action_object)
        return result.err()

    @service_method(path="action.save", name="save")
    def save(
        self,
        context: AuthedServiceContext,
        action_object: Union[ActionObject, TwinObject],
    ) -> Result[SyftSuccess, str]:
        """Save an object to the action store"""
        # 🟡 TODO 9: Create some kind of type checking / protocol for SyftSerializable
        result = self.store.set(
            uid=action_object.id,
            credentials=context.credentials,
            syft_object=action_object,
        )
        if result.is_ok():
            return Ok(SyftSuccess(message=f"{type(action_object)} saved"))
        return result.err()

    @service_method(path="action.get", name="get")
    def get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        skip_permission: bool = False,
    ) -> Result[ActionObject, str]:
        """Get an object from the action store"""
        # TODO 🟣 Temporarily added skip permission arguments for enclave
        # until permissions are fully integrated
        result = self.store.get(
            uid=uid, credentials=context.credentials, skip_permission=skip_permission
        )
        if result.is_ok():
            obj = result.ok()
            if isinstance(obj, TwinObject):
                if twin_mode == TwinMode.PRIVATE:
                    obj = obj.private
                    obj.syft_point_to(context.node.id)
                elif twin_mode == TwinMode.MOCK:
                    obj = obj.mock
                    obj.syft_point_to(context.node.id)
                else:
                    obj.mock.syft_point_to(context.node.id)
                    obj.private.syft_point_to(context.node.id)
            return Ok(obj)
        return Err(result.err())

    @service_method(path="action.get_pointer", name="get_pointer")
    def get_pointer(
        self, context: AuthedServiceContext, uid: UID
    ) -> Result[ActionObjectPointer, str]:
        """Get a pointer from the action store"""
        result = self.store.get_pointer(
            uid=uid, credentials=context.credentials, node_uid=context.node.id
        )
        if result.is_ok():
            return Ok(result.ok())
        return Err(result.err())

    # not a public service endpoint
    def _user_code_execute(
        self, context: AuthedServiceContext, code_item: UserCode, kwargs: Dict[str, Any]
    ) -> Result[ActionObjectPointer, Err]:
        filtered_kwargs = code_item.input_policy.filter_kwargs(
            kwargs=kwargs, context=context, code_item_id=code_item.id
        )
        if filtered_kwargs.is_err():
            return filtered_kwargs
        filtered_kwargs = filtered_kwargs.ok()
        has_twin_inputs = False
        kwargs = {}
        for key, kwarg_value in filtered_kwargs.items():
            if isinstance(kwarg_value, TwinObject):
                has_twin_inputs = True
            kwargs[key] = kwarg_value

        result_id = UID()

        try:
            if not has_twin_inputs:
                # no twins
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.NONE)
                exec_result = execute_byte_code(code_item, filtered_kwargs)
                result_action_object = wrap_result(
                    code_item.id, result_id, exec_result.result
                )
            else:
                # twins
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.PRIVATE)
                private_exec_result = execute_byte_code(code_item, private_kwargs)
                result_action_object_private = wrap_result(
                    code_item.id, result_id, private_exec_result.result
                )

                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.MOCK)
                mock_exec_result = execute_byte_code(code_item, mock_kwargs)
                result_action_object_mock = wrap_result(
                    code_item.id, result_id, mock_exec_result.result
                )

                result_action_object = TwinObject(
                    id=result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
        except Exception as e:
            print("what is this exception", e)
            return Err("_user_code_execute failed")

        set_result = self.store.set(
            uid=result_id,
            credentials=context.credentials,
            syft_object=result_action_object,
        )
        if set_result.is_err():
            return set_result.err()

        if isinstance(result_action_object, TwinObject):
            result_action_object = result_action_object.mock
        result_action_object.syft_point_to(context.node.id)

        return Ok(result_action_object)

    @service_method(path="action.execute", name="execute")
    def execute(
        self, context: AuthedServiceContext, action: Action
    ) -> Result[ActionObjectPointer, Err]:
        """Execute an operation on objects in the action store"""
        resolved_self = self.get(
            context=context, uid=action.remote_self, twin_mode=TwinMode.NONE
        )
        if resolved_self.is_err():
            return resolved_self.err()
        resolved_self = resolved_self.ok()

        if isinstance(resolved_self, TwinObject):
            private_result = execute_object(
                self, context, resolved_self.private, action, twin_mode=TwinMode.PRIVATE
            )
            if private_result.is_err():
                return private_result.err()
            mock_result = execute_object(
                self, context, resolved_self.mock, action, twin_mode=TwinMode.MOCK
            )
            if mock_result.is_err():
                return mock_result.err()

            private_result = private_result.ok()
            mock_result = mock_result.ok()

            result_action_object = Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=private_result,
                    private_obj_id=action.result_id,
                    mock_obj=mock_result,
                    mock_obj_id=action.result_id,
                )
            )
        else:
            result_action_object = execute_object(self, context, resolved_self, action)

        if result_action_object.is_err():
            return result_action_object.err()
        else:
            result_action_object = result_action_object.ok()

        set_result = self.store.set(
            uid=action.result_id,
            credentials=context.credentials,
            syft_object=result_action_object,
        )
        if set_result.is_err():
            return set_result.err()

        if isinstance(result_action_object, TwinObject):
            result_action_object = result_action_object.mock
        result_action_object.syft_point_to(context.node.id)

        return Ok(result_action_object)

    @service_method(path="action.exists", name="exists")
    def exists(
        self, context: AuthedServiceContext, obj_id: UID
    ) -> Result[SyftSuccess, SyftError]:
        """Checks if the given object id exists in the Action Store"""
        if self.store.exists(obj_id):
            return SyftSuccess(message=f"Object: {obj_id} exists")
        else:
            return SyftError(message=f"Object: {obj_id} does not exist")


def execute_object(
    service: ActionService,
    context: AuthedServiceContext,
    resolved_self: ActionObject,
    action: Action,
    twin_mode: TwinMode = TwinMode.NONE,
) -> Result[Union[TwinObject, ActionObject], str]:
    resolved_self = resolved_self.syft_action_data
    args = []
    has_twin_inputs = False
    if action.args:
        for arg_id in action.args:
            arg_value = service.get(
                context=context, uid=arg_id, twin_mode=TwinMode.NONE
            )
            if arg_value.is_err():
                return arg_value.err()
            if isinstance(arg_value.ok(), TwinObject):
                has_twin_inputs = True
            args.append(arg_value.ok())

    kwargs = {}
    if action.kwargs:
        for key, arg_id in action.kwargs.items():
            kwarg_value = service.get(
                context=context, uid=arg_id, twin_mode=TwinMode.NONE
            )
            if kwarg_value.is_err():
                return kwarg_value.err()
            if isinstance(kwarg_value.ok(), TwinObject):
                has_twin_inputs = True
            kwargs[key] = kwarg_value.ok()

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    target_method = getattr(resolved_self, action.op, None)
    result = None
    try:
        if target_method:
            if twin_mode == TwinMode.NONE and not has_twin_inputs:
                # no twins
                filtered_args = filter_twin_args(args, twin_mode=twin_mode)
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*filtered_args, **filtered_kwargs)
                result_action_object = wrap_result(
                    action.parent_id, action.result_id, result
                )
            elif twin_mode == TwinMode.NONE and has_twin_inputs:
                # self isn't a twin but one of the inputs is
                private_args = filter_twin_args(args, twin_mode=twin_mode)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                private_result = target_method(*private_args, **private_kwargs)
                result_action_object_private = wrap_result(
                    action.parent_id, action.result_id, private_result
                )

                mock_args = filter_twin_args(args, twin_mode=twin_mode)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                mock_result = target_method(*mock_args, **mock_kwargs)
                result_action_object_mock = wrap_result(
                    action.parent_id, action.result_id, mock_result
                )

                result_action_object = TwinObject(
                    id=action.result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
            elif twin_mode == twin_mode.PRIVATE:
                # twin private path
                private_args = filter_twin_args(args, twin_mode=twin_mode)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*private_args, **private_kwargs)
                result_action_object = wrap_result(
                    action.parent_id, action.result_id, result
                )
            elif twin_mode == twin_mode.MOCK:
                # twin mock path
                mock_args = filter_twin_args(args, twin_mode=twin_mode)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                target_method = getattr(resolved_self, action.op, None)
                result = target_method(*mock_args, **mock_kwargs)
                result_action_object = wrap_result(
                    action.parent_id, action.result_id, result
                )
            else:
                raise Exception(
                    f"Bad combination of: twin_mode: {twin_mode} and has_twin_inputs: {has_twin_inputs}"
                )

    except Exception as e:
        print("what is this exception", e)
        return Err(e)
    return Ok(result_action_object)


def wrap_result(parent_id: UID, result_id: UID, result: Any) -> ActionObject:
    # 🟡 TODO 11: Figure out how we want to store action object results
    action_type = action_type_for_type(result)
    if action_type is None:
        print("action_type_for_type(result)", action_type_for_type(result))
        raise Exception(f"No Action Type for type: {type(result)}")
    result_action_object = action_type(
        id=result_id, parent_id=parent_id, syft_action_data=result
    )
    return result_action_object


def filter_twin_args(args: List[Any], twin_mode: TwinMode) -> Any:
    filtered = []
    for arg in args:
        if isinstance(arg, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered.append(arg.private.syft_action_data)
            elif twin_mode == TwinMode.MOCK:
                filtered.append(arg.mock.syft_action_data)
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered.append(arg.syft_action_data)
    return filtered


def filter_twin_kwargs(kwargs: Dict, twin_mode: TwinMode) -> Any:
    filtered = {}
    for k, v in kwargs.items():
        if isinstance(v, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered[k] = v.private.syft_action_data
            elif twin_mode == TwinMode.MOCK:
                filtered[k] = v.mock.syft_action_data
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered[k] = v.syft_action_data
    return filtered


TYPE_TO_SERVICE[ActionObject] = ActionService
TYPE_TO_SERVICE[TwinObject] = ActionService
TYPE_TO_SERVICE[AnyActionObject] = ActionService

SERVICE_TO_TYPES[ActionService].update({ActionObject, TwinObject, AnyActionObject})



# stdlib
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# relative
from ....telemetry import instrument
from .context import AuthedServiceContext
from .context import NodeServiceContext
from .context import UnauthedServiceContext
from .credentials import SyftVerifyKey
from .credentials import UserLoginCredentials
from .document_store import DocumentStore
from .response import SyftError
from .response import SyftSuccess
from .serializable import serializable
from .service import AbstractService
from .service import SERVICE_TO_TYPES
from .service import TYPE_TO_SERVICE
from .service import service_method
from .uid import UID
from .user import ServiceRole
from .user import User
from .user import UserCreate
from .user import UserPrivateKey
from .user import UserSearch
from .user import UserUpdate
from .user import UserView
from .user import check_pwd
from .user import salt_and_hash_password
from .user_stash import UserStash


@instrument
@serializable(recursive_serde=True)
class UserService(AbstractService):
    store: DocumentStore
    stash: UserStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = UserStash(store=store)

    @service_method(path="user.create", name="create")
    def create(
        self, context: AuthedServiceContext, user_create: UserCreate
    ) -> Union[UserView, SyftError]:
        """Create a new user"""
        user = user_create.to(User)
        result = self.stash.get_by_email(email=user.email)
        if result.is_err():
            return SyftError(message=str(result.err()))
        user_exists = result.ok() is not None
        if user_exists:
            return SyftError(message=f"User already exists with email: {user.email}")

        result = self.stash.set(user=user)
        if result.is_err():
            return SyftError(message=str(result.err()))
        user = result.ok()
        return user.to(UserView)

    @service_method(path="user.view", name="view")
    def view(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[Optional[UserView], SyftError]:
        """Get user for given uid"""
        result = self.stash.get_by_uid(uid=uid)
        if result.is_ok():
            user = result.ok()
            if user is None:
                return SyftError(message=f"No user exists for given: {uid}")
            return user.to(UserView)

        return SyftError(message=str(result.err()))

    @service_method(path="user.get_all", name="get_all")
    def get_all(
        self, context: AuthedServiceContext
    ) -> Union[Optional[UserView], SyftError]:
        result = self.stash.get_all()
        if result.is_ok():
            return result.ok()

        # 🟡 TODO: No user exists will happen when result.ok() is empty list
        return SyftError(message="No users exists")

    @service_method(path="user.search", name="search", autosplat=["user_search"])
    def search(
        self,
        context: AuthedServiceContext,
        user_search: UserSearch,
    ) -> Union[List[UserView], SyftError]:
        kwargs = user_search.to_dict(exclude_none=True)

        if len(kwargs) == 0:
            valid_search_params = list(UserSearch.__fields__.keys())
            return SyftError(
                message=f"Invalid Search parameters. \
                Allowed params: {valid_search_params}"
            )
        result = self.stash.find_all(**kwargs)
        if result.is_err():
            return SyftError(message=str(result.err()))
        users = result.ok()
        return [user.to(UserView) for user in users] if users is not None else []

    @service_method(path="user.update", name="update")
    def update(
        self, context: AuthedServiceContext, uid: UID, user_update: UserUpdate
    ) -> Union[UserView, SyftError]:
        # TODO: ADD Email Validation

        # Get user to be updated by its UID
        result = self.stash.get_by_uid(uid=uid)

        if result.is_err():
            error_msg = (
                f"Failed to find user with UID: {uid}. Error: {str(result.err())}"
            )
            return SyftError(message=error_msg)

        user = result.ok()

        if user is None:
            return SyftError(message=f"No user exists for given UID: {uid}")

        # Fill User Update fields that will not be changed by replacing it
        # for the current values found in user obj.
        for name, value in vars(user_update).items():
            if name == "password" and value:
                salt, hashed = salt_and_hash_password(value, 12)
                user.hashed_password = hashed
                user.salt = salt
            elif not name.startswith("__") and value is not None:
                setattr(user, name, value)

        result = self.stash.update(user=user)

        if result.is_err():
            error_msg = (
                f"Failed to update user with UID: {uid}. Error: {str(result.err())}"
            )
            return SyftError(message=error_msg)

        user = result.ok()

        return user.to(UserView)

    @service_method(path="user.delete", name="delete")
    def delete(self, context: AuthedServiceContext, uid: UID) -> Union[bool, SyftError]:
        result = self.stash.delete_by_uid(uid=uid)
        if result.is_err():
            return SyftError(message=str(result.err()))

        return result.ok()

    def exchange_credentials(
        self, context: UnauthedServiceContext
    ) -> Union[UserLoginCredentials, SyftError]:
        """Verify user
        TODO: We might want to use a SyftObject instead
        """
        result = self.stash.get_by_email(email=context.login_credentials.email)
        if result.is_ok():
            user = result.ok()
            if user is not None and check_pwd(
                context.login_credentials.password,
                user.hashed_password,
            ):
                return user.to(UserPrivateKey)

            return SyftError(
                message="No user exists with "
                f"{context.login_credentials.email} and supplied password."
            )

        return SyftError(
            message="Failed to retrieve user with "
            f"{context.login_credentials.email} with error: {result.err()}"
        )

    def admin_verify_key(self) -> Union[SyftVerifyKey, SyftError]:
        result = self.stash.get_by_role(role=ServiceRole.ADMIN)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok().verify_key

    def register(
        self, context: NodeServiceContext, new_user: UserCreate
    ) -> Union[Tuple[SyftSuccess, UserPrivateKey], SyftError]:
        """Register new user"""

        user = new_user.to(User)
        result = self.stash.get_by_email(email=user.email)
        if result.is_err():
            return SyftError(message=str(result.err()))
        user_exists = result.ok() is not None
        if user_exists:
            return SyftError(message=f"User already exists with email: {user.email}")

        result = self.stash.set(user=user)
        if result.is_err():
            return SyftError(message=str(result.err()))

        user = result.ok()
        msg = SyftSuccess(message=f"{user.email} User successfully registered !!!")
        return tuple([msg, user.to(UserPrivateKey)])

    def user_verify_key(self, email: str) -> Union[SyftVerifyKey, SyftError]:
        result = self.stash.get_by_email(email=email)
        if result.is_ok():
            return result.ok().verify_key
        return SyftError(message=f"No user with email: {email}")


TYPE_TO_SERVICE[User] = UserService
SERVICE_TO_TYPES[UserService].update({User})



# stdlib
from typing import Any

# syft absolute
import syft

# relative
from .recursive import recursive_serde_register

module_type = type(syft)


def serializable(
    recursive_serde: bool = False,
) -> Any:
    def rs_decorator(cls: Any) -> Any:
        recursive_serde_register(cls)
        return cls

    if recursive_serde:
        return rs_decorator



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional

# relative
from .action_object import ActionObject
from .action_types import action_types
from .serializable import serializable
from .syft_object import SyftObject
from .uid import UID


def to_action_object(obj: Any) -> ActionObject:
    if isinstance(obj, ActionObject):
        return obj

    if type(obj) in action_types:
        return action_types[type(obj)](syft_action_data=obj)
    raise Exception(f"{type(obj)} not in action_types")


@serializable(recursive_serde=True)
class TwinObject(SyftObject):
    __canonical_name__ = "TwinObject"
    __version__ = 1

    __attr_searchable__ = []
    __attr_state__ = ["id", "private_obj", "private_obj_id", "mock_obj", "mock_obj_id"]

    private_obj: ActionObject
    private_obj_id: UID
    mock_obj: ActionObject
    mock_obj_id: UID

    def __init__(
        self,
        private_obj: ActionObject,
        mock_obj: ActionObject,
        private_obj_id: Optional[UID] = None,
        mock_obj_id: Optional[UID] = None,
        id: Optional[UID] = None,
    ) -> None:
        private_obj = to_action_object(private_obj)
        mock_obj = to_action_object(mock_obj)

        if private_obj_id is None:
            private_obj_id = private_obj.id
        if mock_obj_id is None:
            mock_obj_id = mock_obj.id
        if id is None:
            id = UID()
        super().__init__(
            private_obj=private_obj,
            private_obj_id=private_obj_id,
            mock_obj=mock_obj,
            mock_obj_id=mock_obj_id,
            id=id,
        )

    @property
    def private(self) -> ActionObject:
        twin_id = self.id
        private = self.private_obj
        private.id = twin_id
        return private

    @property
    def mock(self) -> ActionObject:
        twin_id = self.id
        mock = self.mock_obj
        mock.id = twin_id
        return mock



# stdlib
from typing import List

# third party
from result import Result

# relative
from ....telemetry import instrument
from .document_store import BaseUIDStoreStash
from .document_store import DocumentStore
from .document_store import PartitionKey
from .document_store import PartitionSettings
from .node_metadata import NodeMetadata
from .serializable import serializable
from .uid import UID

NamePartitionKey = PartitionKey(key="name", type_=str)
ActionIDsPartitionKey = PartitionKey(key="action_ids", type_=List[UID])


@instrument
@serializable(recursive_serde=True)
class MetadataStash(BaseUIDStoreStash):
    object_type = NodeMetadata
    settings: PartitionSettings = PartitionSettings(
        name=NodeMetadata.__canonical_name__, object_type=NodeMetadata
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set(self, metadata: NodeMetadata) -> Result[NodeMetadata, str]:
        return self.check_type(metadata, self.object_type).and_then(super().set)

    def update(self, metadata: NodeMetadata) -> Result[NodeMetadata, str]:
        return self.check_type(metadata, self.object_type).and_then(super().update)



# relative
from .base import SyftBaseModel
from .serializable import serializable


class SyftResponseMessage(SyftBaseModel):
    message: str
    _bool: bool = True

    def __bool__(self) -> bool:
        return self._bool

    def __eq__(self, other) -> bool:
        if isinstance(other, SyftResponseMessage):
            return self.message == other.message and self._bool == other._bool
        return self._bool == other

    def __repr__(self) -> str:
        return f"{type(self)}: {self.message}"

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}" style="padding:5px;">'
            + f"<strong>{type(self).__name__}</strong>: {self.message}</div><br />"
        )


@serializable(recursive_serde=True)
class SyftError(SyftResponseMessage):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"


@serializable(recursive_serde=True)
class SyftSuccess(SyftResponseMessage):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-success"


@serializable(recursive_serde=True)
class SyftNotReady(SyftResponseMessage):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"


@serializable(recursive_serde=True)
class SyftException(Exception):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}" style="padding:5px;">'
            + f"<strong>{type(self).__name__}</strong>: {self.args}</div><br />"
        )



# stdlib
from typing import Callable
from typing import List
from typing import Union

# third party
import numpy as np
import pandas as pd

# relative
from .action_object import ActionObject
from .response import SyftError
from .response import SyftResponseMessage
from .response import SyftSuccess


def verify_result(
    func: Callable,
    private_inputs: Union[ActionObject, List[ActionObject]],
    private_outputs: Union[ActionObject, List[ActionObject]],
) -> SyftResponseMessage:
    """Verify a single result of Code Verification"""
    trace_assets = []
    if not isinstance(private_inputs, List):
        private_inputs = [private_inputs]

    for asset in private_inputs:
        if not isinstance(asset, ActionObject):
            raise Exception(f"ActionObject expected, instead received: {type(asset)}")
        # Manual type casting for now, to automate later
        if isinstance(asset.syft_action_data, np.ndarray):
            empty_obj: ActionObject = ActionObject(
                id=asset.id, syft_result_obj=np.ndarray([])
            )
        elif isinstance(asset.syft_action_data, pd.DataFrame):
            empty_obj: ActionObject = ActionObject(
                id=asset.id, syft_result_obj=pd.DataFrame()
            )
        else:
            raise NotImplementedError(
                f"Trace mode not yet automated for type: {type(asset.syft_action_data)}"
            )

        trace_assets.append(empty_obj)

    print("Code Verification in progress.")
    traced_results = func(*trace_assets)

    if isinstance(private_outputs, List):
        target_hashes = [output.syft_history_hash for output in private_outputs]
        traced_hashes = [result.syft_history_hash for result in traced_results]
    else:
        target_hashes = private_outputs.syft_history_hash
        traced_hashes = traced_results.syft_history_hash

    if target_hashes == traced_hashes:
        msg = "Code Verification passed with matching hashes! Congratulations, and thank you for supporting PySyft!"
        return SyftSuccess(message=msg)
    else:
        msg = (
            f"Hashes do not match! Target hashes were: {target_hashes} but Traced hashes were: {traced_results}. "
            f"Please try checking the logs."
        )
        return SyftError(message=msg)


def code_verification(func: Callable):
    """Compares history hashes of an Empty Action Object to that of the real action object

    Inputs:
    - func:: a Callable whose sole argument should be the Private Dataset(s) being used. Constraints:
        - Input arguments are the private datasets being used
        - Output arguments are the results requested.

    Outputs:
    - boolean:: if history hashes match
    """

    def wrapper(*args, **kwargs):
        trace_assets = []
        for asset in args:
            if not isinstance(asset, ActionObject):
                raise Exception(
                    f"ActionObject expected, instead received: {type(asset)}"
                )
            # Manual type casting for now, to automate later
            if isinstance(asset.syft_action_data, np.ndarray):
                empty_obj: ActionObject = ActionObject(
                    id=asset.id, syft_result_obj=np.ndarray([])
                )
            elif isinstance(asset.syft_action_data, pd.DataFrame):
                empty_obj: ActionObject = ActionObject(
                    id=asset.id, syft_result_obj=pd.DataFrame()
                )
            else:
                raise NotImplementedError(
                    f"Trace mode not yet automated for type: {type(asset.syft_action_data)}"
                )

            trace_assets.append(empty_obj)

        print("Evaluating function normally to obtain history hash")
        results = func(*args, **kwargs).syft_history_hash
        print(8 * "(-(-_(-_-)_-)-)")

        print("Tracing function to obtain history hash")
        traced_results = func(*trace_assets, **kwargs).syft_history_hash
        print(8 * "(-(-_(-_-)_-)-)")

        # assert len(results) == len(traced_results)
        hashes_match = results == traced_results
        if hashes_match:
            msg = (
                f"Code Verification passed with matching hashes of {results}! Congratulations, and thank you for "
                f"supporting PySyft!"
            )
            return SyftSuccess(message=msg)
        else:
            msg = (
                f"Hashes do not match! Target hashes were: {results} but Traced hashes were: {traced_results}. "
                f"Please try checking the logs."
            )
            return SyftError(message=msg)

    return wrapper
